/*------------------------------------------------------------------------*/

#define COMPSAT_COPYRIGHT \
  "(C)2002-2005 Armin Biere, ETH Zurich and JKU Linz"

#define COMPSAT_ID \
"$Id: compsat.c,v 1.78 2005/10/01 13:27:50 biere Exp $"

/*------------------------------------------------------------------------*/

#include "config.h"
#include "compsat.h"
#include "dimacs.h"

/*------------------------------------------------------------------------*/
#ifndef COMPSAT_CHECK
#define NDEBUG
#endif
#include <assert.h>
/*------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <ctype.h>

/*------------------------------------------------------------------------*/

#define COMPSAT_PREFIX		"c compsat "

#define WATCHED			2
#define ISORT_LIMIT		10

#define LD_BITS_PER_WORD	5
#define BITS_PER_WORD		(1 << LD_BITS_PER_WORD)

#define FALSE			-1
#define TRUE			1

#define MAX			(1 << 28)

#define GARBAGE			INT_MIN

/*------------------------------------------------------------------------*/
/* There are four types of conflicts/reasons/assignments:
 *
 *  1 <= INT_MAX        index into literal data base
 *  0			no conflict (invalid reason/assignment)
 *
 *  -2*MAX-1,...,-MAX-1,...,-1  	unit reason/assignment/conflict
 *     (-MAX,...,     0,...,MAX)
 *
 *  -4*MAX-2,...,-3*MAX-2,...,-2*Max-2 	decision reason/assignment/conflict
 *     (-MAX,...,     0,...,MAX)
 */
enum Reason
{
  MIN_UNIT = -2*MAX-1,
  UNIT_OFFSET = -MAX-1,
  DECISION_OFFSET = -3*MAX-2,
};

/*------------------------------------------------------------------------*/

typedef struct Counter Counter;
typedef struct Counters Counters;
typedef struct Limits Limits;
typedef struct Stack Stack;
typedef struct Statistics Statistics;
typedef struct Variable Variable;

/*------------------------------------------------------------------------*/

struct Stack
{
  int *start, *top, *end;
};

/*------------------------------------------------------------------------*/

struct Counter
{
  int current, init, limit, max, factor;
  double finished;
};

/*------------------------------------------------------------------------*/

struct Counters
{
  Counter restart;
  Counter rescore;
  Counter report;
  Counter sort;			
};

/*------------------------------------------------------------------------*/

struct Statistics
{
  size_t max_bytes;		/* maximal number of allocated bytes */
  size_t bytes;			/* number of currently allocated bytes */
  double time;			/* accumulative seconds spent in library */
  int iterations;
#ifdef DECOMPOSE
  int decompositions;
  int switches;
  int components;
  int non_trivial_components;
  int max_non_trivial_components;
  int saved;
  double decomposition_time;
  int cross_component_collected_learned_clauses;
#endif
  int gcs;
  double assignments;
  int toplevel;
  double conflicts;
  double decisions;
  double osearched;
  double lsearched;
  double csearched;
  double sum_levels_at_decision;
  double collected_literals;
  double collected_implications;
  double collected_clauses;
  unsigned reported;
  int pure_literals;		/* toplevel */
  int zombies;
  int units;			/* toplevel */
  int variables;		/* number of occuring variables */
  int implications;
  int clauses;
  int literals;
};

/*------------------------------------------------------------------------*/

struct Limits
{
  double bytes;			/* byte limit (default -1 unrestricted) */
  double decisions;		/* decisions+1 (default -1 unrestricted) */
  double time;			/* time limit (default -1 unrestricted) */
};

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/
/* If checking is enabled we can check that a conflict clause to be added is
 * actually implied by the current clause set.  This check is performed by
 * unit resolution with a seperate simple 'Checker' solver.
 */
typedef struct Checker Checker;

struct Checker
{
  int enabled;			/* dynamically */
  CompSat *compsat;
  Stack literals;
  int * assignment;		/* our own assignment */
  Stack *occurence[2];		/* literal occurence in compsat->literals */
  Stack stack;			/* the pending assignment DFS stack */
  Stack units;			/* the unit clauses */
};

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

struct Variable
{
  Stack watched[2];
  Stack implies[2];
  int level;			/* level where this variable was assigned */
  int reason;			/* reason for this assignment */
#ifdef DECOMPOSE
  int component;
  int component_size;
#endif
  int pos[2];			/* position of literals in 'order' */
  int score[2];			/* score for each literal */
  int ref[2];			/* number of occurences */
  unsigned salt[2];		/* some random 'salt' for ordering */
};

/*------------------------------------------------------------------------*/

struct CompSat
{
  signed char *assignment;
  Variable *variables;

  Stack literals;		/* contains all literals and clauses */
  int level;

  Stack units;			/* indices of unit clauses */

  Stack mark[2];		/* bit map for mark flags of literals */
  Stack used;			/* bit map of used variables */

  int next_implies;
  int next_watched;
  Stack trail;			/* the indices of assigned variables */
  Stack control;		/* maps decision to trail level */
  Stack order;			/* orders the literals */
  Stack clause;			/* for conflicts and simplifying */
  int unassigned;		/* first potentially unassigned variable */
  int zero;			/* first necessarilly zero score variable */
  Stack visited;		/* save visited literals */
  Stack stack;			/* multipurpose stack */
  Stack added_literals;		/* only used by 'compsat_add_literal' */
  int max;			/* maximal variable id */
  int size;			/* size of watched arrays or variables */
  int learned;			/* index of first learned clause */
  int satisfied;		/* cached satisfied learned clause */
  int iteration;		/* index of first clause of this iteration */
#ifdef DECOMPOSE
  int active;			/* index of active component */
#endif
  int contains_empty_clause;
  int restart_at_next_decision;
  int conflict;

  FILE *log;			/* log file */
  int close_log;
  FILE *out;			/* output file */
  int close_out;
  int divert;			/* divert logging to output */
  int verbosity;		/* verbosity level for logging */
  int logged_verbosity;
  int no_output;		/* disable all output (except logging) */

  char *error;			/* contains error message */
  char *name;			/* file name */

  int internal;			/* number of nested internal invocations */
  double entered;		/* enter time */

  Limits limits;
  Statistics stats;
  Counters counters;

  unsigned rng;

#ifdef COMPSAT_CHECK
  size_t *leaked_ptr;		/* report leaked bytes */
  Checker checker;		/* e.g. conflict clauses implication */
#endif
#ifdef COMPSAT_WHITE
  int force_out_of_memory;	/* see 'memout' white box test */
#endif
};

/*------------------------------------------------------------------------*/

struct CompSatArgvIteratorInternal
{
  CompSat *compsat;
  int argc;			/* size of 'argv' ('argv[0]' is skipped) */
  char **argv;			/* copy of original 'argv' argument */
  int pos;			/* current position in 'argv' */
  char opt_buffer[3];		/* holds '-t' if '-t1.2' specified etc. */
};

/*------------------------------------------------------------------------*/

inline_static void *
new (CompSat * compsat, size_t size)
{
  size_t *res, bytes;

  if (!size)
    return 0;

  bytes = size;
#ifdef COMPSAT_CHECK
  bytes += sizeof (size_t);
#endif

#ifdef COMPSAT_WHITE
  if (compsat->force_out_of_memory)
    res = 0;
  else
#endif
    res = (size_t *) malloc (bytes);

  if (!res)
    return 0;

#ifdef COMPSAT_CHECK
  *res++ = size;
#endif
  memset (res, 0, size);
  compsat->stats.bytes += size;
  if (compsat->stats.max_bytes < compsat->stats.bytes)
    compsat->stats.max_bytes = compsat->stats.bytes;

  return (void *) res;
}

/*------------------------------------------------------------------------*/

inline_static void
delete (CompSat * compsat, void *ptr, size_t size)
{
  size_t *data;

  if (!ptr)
    {
      assert (!size);
      return;
    }

#ifdef COMPSAT_CHECK
  data = ((size_t *) ptr) - 1;
  assert (*data == size);
#else
  data = ptr;
  (void) size;
#endif
  assert (compsat->stats.bytes >= size);
  compsat->stats.bytes -= size;
  free ((void *) data);
}

/*-----------------------------------------------------------------------*/

inline_static void *
resize (CompSat * compsat, void *ptr, size_t old_size, size_t new_size)
{
  size_t *data, bytes;

  if (!ptr)
    {
      assert (!old_size);
      return new (compsat, new_size);
    }

  if (!new_size)
    {
      delete (compsat, ptr, old_size);
      return 0;
    }

#ifdef COMPSAT_CHECK
  bytes = new_size + sizeof (size_t);
  data = ((size_t *) ptr) - 1;
  assert (*data == old_size);
#else
  bytes = new_size;
  data = ptr;
#endif

#ifdef COMPSAT_WHITE
  if (compsat->force_out_of_memory)
    data = 0;
  else
#endif
    data = realloc (data, bytes);

  if (!data)
    return 0;

#ifdef COMPSAT_CHECK
  *data++ = new_size;
#endif

  if (new_size > old_size)
    memset (old_size + (char *) data, 0, new_size - old_size);

  assert (compsat->stats.bytes >= old_size);
  compsat->stats.bytes -= old_size;
  compsat->stats.bytes += new_size;
  if (compsat->stats.max_bytes < compsat->stats.bytes)
    compsat->stats.max_bytes = compsat->stats.bytes;

  return data;
}

/*------------------------------------------------------------------------*/

inline_static char *
copy_string (CompSat * compsat, const char *str)
{
  char *res;

  res = 0;

  if (str)
    {
      res = (char *) new (compsat, strlen (str) + 1);
      if (res)
	strcpy (res, str);
    }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
delete_string (CompSat * compsat, char *str)
{
  if (str)
    delete (compsat, str, strlen (str) + 1);
}

/*------------------------------------------------------------------------*/

double
compsat_get_time (void)
{
  struct rusage u;
  double res;

  res = 0;

  if (!getrusage (RUSAGE_SELF, &u))
    {
      res += u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
      res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
    }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static double
delta_time (CompSat * compsat)
{
  double res;

  assert (compsat->internal);

  res = compsat_get_time () - compsat->entered;
  if (res <= 0)
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
update_time (CompSat * compsat)
{
  double current, delta;

  assert (compsat->internal);

  current = compsat_get_time ();
  delta = current - compsat->entered;
  delta = (delta < 0) ? 0 : delta;
  compsat->stats.time += delta;
  compsat->entered = current;
}

/*------------------------------------------------------------------------*/

inline_static void
reset_counter (Counter * counter)
{
  counter->current = 0;
}

/*------------------------------------------------------------------------*/

inline_static void
init_counter (Counter * counter, int init, int factor, int max)
{
  assert (init >= 0);

  assert (max < 0 || init <= max);

  counter->current = 0;
  counter->init = init;
  counter->limit = counter->init;
  counter->max = max;

  counter->factor = factor;
  counter->finished = 0;
}

/*------------------------------------------------------------------------*/

inline_static int
done_counter (Counter * counter)
{
  int res;

  if (counter->init)
    {
      if (counter->current < counter->limit)
	{
	  counter->current++;
	  res = 0;
	}
      else
	{
	  counter->init *= counter->factor;
	  if (counter->max >= 0 && counter->init > counter->max)
	    counter->init = counter->max;
	  counter->limit = counter->init;
	  counter->current = 0;
	  counter->finished++;
	  res = 1;
	}
    }
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/

inline_static void
check_stack (Stack * stack)
{
  assert (stack->start <= stack->top);
  assert (stack->top <= stack->end);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static size_t
bytes_stack (Stack * stack)
{
  return ((char *) stack->end) - ((char *) stack->start);
}

/*------------------------------------------------------------------------*/

inline_static int
size_stack (Stack * stack)
{
  return stack->end - stack->start;
}

/*------------------------------------------------------------------------*/

inline_static int
count_stack (Stack * stack)
{
  return stack->top - stack->start;
}

/*------------------------------------------------------------------------*/

inline_static void
reset_stack (CompSat * compsat, Stack * stack, int count)
{
  assert (count >= 0);
  assert (count <= count_stack (stack));
  stack->top = stack->start + count;
#ifdef COMPSAT_CHECK
  check_stack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
release_stack (CompSat * compsat, Stack * stack)
{
  delete (compsat, stack->start, bytes_stack (stack));
  memset (stack, 0, sizeof (*stack));
#ifdef COMPSAT_CHECK
  check_stack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
enlarge_stack (CompSat * compsat, Stack * stack)
{
  size_t old_bytes, new_bytes;
  int *old_start, *new_start;
  ptrdiff_t delta;

  old_bytes = bytes_stack (stack);
  new_bytes = old_bytes ? 2 * old_bytes : sizeof (int);

  old_start = stack->start;
  new_start = (int *) resize (compsat, old_start, old_bytes, new_bytes);
  delta = new_start - old_start;

  stack->start = new_start;
  stack->top += delta;
  stack->end = (int *) (new_bytes + (char *) stack->start);

#ifdef COMPSAT_CHECK
  check_stack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
push_stack (CompSat * compsat, Stack * stack, int n)
{
  assert (n != INT_MIN);

  if (stack->top == stack->end)
    enlarge_stack (compsat, stack);

  *stack->top++ = n;

#ifdef COMPSAT_CHECK
  check_stack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
is_empty_stack (Stack * stack)
{
  return stack->top == stack->start;
}

/*------------------------------------------------------------------------*/

inline_static int
pop_stack (CompSat * compsat, Stack * stack)
{
  int res;

  assert (!is_empty_stack (stack));
  res = *--stack->top;
#ifdef COMPSAT_CHECK
  check_stack (stack);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
set_stack (CompSat * compsat, Stack * stack, int idx, int val)
{
  int *p;

  assert (idx >= 0);

  while (size_stack (stack) <= idx)
    enlarge_stack (compsat, stack);

  p = stack->start + idx;
  if (stack->top <= p)
    {
      memset (stack->top, 0, (char *) p - (char *) stack->top);
      stack->top = p + 1;
    }

  *p = val;

#ifdef COMPSAT_CHECK
  check_stack (stack);
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
get_stack (Stack * stack, int idx)
{
  int *p, res;

  assert (0 <= idx);
  p = stack->start + idx;
  assert (p < stack->top);
  res = *p;

  return res;
}

/*------------------------------------------------------------------------*/

static int
next_stack_aux (int **ptr, const int delta)
{
  int res;

  res = **ptr;
  *ptr += delta;

  return res;
}

/*------------------------------------------------------------------------*/

#define IteratorStack(stack, it, dir) \
  int * it ## _start = (stack)->start; \
  int * it ## _top = (stack)->top; \
  const int it ## _inc = (dir); \
  const int * it ## _done = \
    ((dir) < 0) ? it ## _start - 1 : it ## _top; \
  int * it ## _cursor = \
    ((dir) < 0) ? it ## _top - 1 : it ## _start

/*------------------------------------------------------------------------*/

#define CursorStack(stack, it, pos, dir) \
  int * it ## _start = (stack)->start; \
  const int it ## _inc = (dir); \
  int * it ## _cursor = it ## _start + (pos)

/*------------------------------------------------------------------------*/

#define save_stack_cursor(it, saved_ptr) \
do { (saved_ptr) = it ## _cursor; } while(0)

/*------------------------------------------------------------------------*/

#define restore_stack_cursor(stack, saved_ptr) \
  do { it ## _cursor = (saved_ptr); } while(0)

/*------------------------------------------------------------------------*/

#define swap_stack(stack, it, a, b) \
do { \
  int it ## _tmp; \
  assert ((stack)->start <= (int*)(a) && (int*)(a) < (stack)->top); \
  assert ((stack)->start <= (int*)(b) && (int*)(b) < (stack)->top); \
  it ## _tmp = *(int*)(a); \
  *(int*)(a) = *(int*)(b); \
  *(int*)(b) = it ## _tmp; \
} while(0)

/*------------------------------------------------------------------------*/

#define done_stack(it) \
  (it ## _cursor == it ## _done)

/*------------------------------------------------------------------------*/

#define next_stack(it) \
  next_stack_aux (&it ## _cursor, it ## _inc)

/*------------------------------------------------------------------------*/

inline_static void
release_variable (CompSat * compsat, Variable * v)
{
  release_stack (compsat, &v->watched[0]);
  release_stack (compsat, &v->watched[1]);
  release_stack (compsat, &v->implies[0]);
  release_stack (compsat, &v->implies[1]);
}

/*------------------------------------------------------------------------*/

inline_static void
release_variables (CompSat * compsat)
{
  size_t bytes;
  int i;

  for (i = 1; i <= compsat->max; i++)
    release_variable (compsat, compsat->variables + i);

  bytes = compsat->size * sizeof (compsat->variables[0]);
  delete (compsat, compsat->variables, bytes);

  compsat->assignment -= compsat->size;
  bytes = 2 * compsat->size * sizeof (compsat->assignment[0]);
  delete (compsat, compsat->assignment, bytes);
}

/*------------------------------------------------------------------------*/

inline_static FILE *
out_file (CompSat * compsat)
{
  FILE *file;

  file = compsat->out;
  if (!file)
    file = stdout;

  return file;
}

/*------------------------------------------------------------------------*/

inline_static FILE *
log_file (CompSat * compsat)
{
  FILE *file;

  if (compsat->divert)
    return out_file (compsat);

  file = compsat->log;
  if (!file)
    file = stdout;

  return file;
}

/*------------------------------------------------------------------------*/

inline_static void
log (CompSat * compsat, int level, const char *fmt, ...)
{
  FILE *file;
  va_list ap;

  if (!compsat->logged_verbosity)
    {
      compsat->logged_verbosity = 1;
      log (compsat, 1, "verbosity %d\n", compsat->verbosity);
    }

#ifdef COMPSAT_CHECK
  char *p;
  p = strchr (fmt, '\n');
  assert (p);
  assert (!strchr (p + 1, '\n'));
#endif

  if (level > compsat->verbosity)
    return;

  file = log_file (compsat);
  fputs (COMPSAT_PREFIX, file);
  va_start (ap, fmt);
  vfprintf (file, fmt, ap);
  va_end (ap);
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

inline_static void
out (CompSat * compsat, const char *fmt, ...)
{
  FILE *file;
  va_list ap;

  if (compsat->no_output)
    return;

  file = out_file (compsat);
  va_start (ap, fmt);
  vfprintf (file, fmt, ap);
  va_end (ap);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
ABS (int lit)
{
  return (lit < 0) ? -lit : lit;
}

/*------------------------------------------------------------------------*/

inline_static int
SGN (int lit)
{
  return (lit < 0);
}

/*------------------------------------------------------------------------*/

inline_static int
is_clause_reason (int reason)
{
  return reason > 0;
}

/*------------------------------------------------------------------------*/

inline_static int
is_unit_reason (int reason)
{
  if (is_clause_reason (reason))
    return 0;

  if (reason < MIN_UNIT)
    return 0;

  return 1;
}

/*------------------------------------------------------------------------*/

inline_static int
unit2reason (int lit)
{
  int res;

  assert (0 < ABS (lit));
  res = lit + UNIT_OFFSET;
  assert (is_unit_reason (res));

  return res;
}

/*------------------------------------------------------------------------*/
#ifndef NDEBUG
/*------------------------------------------------------------------------*/

inline_static int
is_decision_reason (int reason)
{
  if (is_clause_reason (reason))
    return 0;

  if (reason >= MIN_UNIT)
    return 0;

  return 1;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
decision2reason (int lit)
{
  int res;

  assert (0 < ABS (lit));
  res = lit + DECISION_OFFSET;
  assert (is_decision_reason (res));

  return res;
}

/*------------------------------------------------------------------------*/
#ifdef DECOMPOSE
/*------------------------------------------------------------------------*/

inline_static void
inc_component_size (CompSat * compsat, int idx)
{
  assert (idx > 0);
  compsat->variables[idx].component_size++;
}

/*------------------------------------------------------------------------*/

inline_static void
set_component_size (CompSat * compsat, int idx, int new_size)
{
  assert (idx > 0);
  compsat->variables[idx].component_size = new_size;
}

/*------------------------------------------------------------------------*/

inline_static int
get_component_size (CompSat * compsat, int idx)
{
  assert (idx > 0);
  return compsat->variables[idx].component_size;
}

/*------------------------------------------------------------------------*/

inline_static int
get_component (CompSat * compsat, int idx)
{
  assert (idx > 0);
  return compsat->variables[idx].component;
}

/*------------------------------------------------------------------------*/

inline_static void
set_component (CompSat * compsat, int idx, int component)
{
  int * p;

  assert (idx > 0);
  assert (component >= 0);
  assert (idx != component);

  p = &compsat->variables[idx].component;
  if (*p== component)
    return;

#ifdef COMPSAT_LOG
  log (compsat, 6, "DECOMPOSITION %d SET COMPONENT OF %d FROM %d TO %d\n",
       compsat->stats.decompositions, idx, *p, component);
#endif
  *p = component;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
get_level (CompSat * compsat, int idx)
{
  assert (idx > 0);
  return compsat->variables[idx].level;
}

/*------------------------------------------------------------------------*/

inline_static void
set_level (CompSat * compsat, int idx, int level)
{
  assert (idx > 0);
  compsat->variables[idx].level = level;
}

/*------------------------------------------------------------------------*/

inline_static int
reason2lit (CompSat * compsat, int reason)
{
  int res, idx, level;

  assert (reason);

  if (is_clause_reason (reason))
    res = get_stack (&compsat->literals, reason);
  else if (is_unit_reason (reason))
    res = reason - UNIT_OFFSET;
  else
    {
      assert (is_decision_reason (reason));
      res = reason - DECISION_OFFSET;

      idx = ABS (res);
      level = get_level (compsat, idx);

      if (level-- > 0)
	{
	  level = get_stack (&compsat->control, level);
	  if (level < count_stack (&compsat->trail))
	    res = get_stack (&compsat->trail, level);
	}
    }

  return res;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_LOG
/*------------------------------------------------------------------------*/

inline_static int
current_decision (CompSat * compsat)
{
  int res, trail_level;

  assert (!is_empty_stack (&compsat->control));
  trail_level = compsat->control.top[-1];
  assert (trail_level >= 0);
  assert (trail_level < count_stack (&compsat->trail));
  res = compsat->trail.start[trail_level];

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
is_decision (CompSat * compsat, int lit)
{
  int decision, res;

  decision = current_decision (compsat);
  res = (decision == lit);
  assert (decision != -lit);

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
log_reason_with_lit (CompSat * compsat, int level, const char *msg, 
                     int lit, int reason)
{
  FILE * file;

  assert (reason);

  if (compsat->verbosity < level)
    return;

  file = log_file (compsat);
  fprintf (file, COMPSAT_PREFIX "%d %s %d ", compsat->level, msg, lit);
  if (is_clause_reason (reason))
    fprintf (file, "[%d]", reason);
  else if (is_unit_reason (reason))
    fputs ("AS UNIT", file);
  else
    {
      assert (is_decision_reason (reason));

      if (is_decision (compsat, lit))
	fputs ("AS DECISION", file);
      else
	fputs ("IMPLIED BY DECISION", file);
    }

  fputc ('\n', file);
  fflush (file);
}

/*------------------------------------------------------------------------*/

inline_static void
log_reason (CompSat * compsat, int level, const char * msg, int reason)
{
  log_reason_with_lit (compsat, level, msg,
                       reason2lit (compsat, reason), reason);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
split_bit (int bit, int *word_ptr, int *mask_ptr)
{
  assert (sizeof (int) * 8 <= BITS_PER_WORD);
  assert (bit >= 0);
  *word_ptr = bit >> LD_BITS_PER_WORD;
  *mask_ptr = (1 << (bit & (BITS_PER_WORD - 1)));
}

/*------------------------------------------------------------------------*/

inline_static int
get_bit (CompSat * compsat, Stack * stack, int bit)
{
  int mask, idx, word;

  split_bit (bit, &idx, &mask);
  word = (idx >= count_stack (stack)) ? 0 : stack->start[idx];
  return (word & mask) != 0;
}

/*------------------------------------------------------------------------*/

inline_static void
clr_bit (CompSat * compsat, Stack * stack, int bit)
{
  int mask, idx, word;

  split_bit (bit, &idx, &mask);
  word = (idx >= count_stack (stack)) ? 0 : stack->start[idx];
  word &= ~mask;
  set_stack (compsat, stack, idx, word);
}

/*------------------------------------------------------------------------*/

inline_static void
set_bit (CompSat * compsat, Stack * stack, int bit)
{
  int mask, idx, word;

  split_bit (bit, &idx, &mask);
  word = (idx >= count_stack (stack)) ? 0 : stack->start[idx];
  word |= mask;
  set_stack (compsat, stack, idx, word);
}

/*------------------------------------------------------------------------*/

inline_static int
get_mark (CompSat * compsat, int lit)
{
  return get_bit (compsat, &compsat->mark[SGN (lit)], ABS (lit));
}

/*------------------------------------------------------------------------*/

inline_static void
set_mark (CompSat * compsat, int lit)
{
  set_bit (compsat, &compsat->mark[SGN (lit)], ABS (lit));
}

/*------------------------------------------------------------------------*/

inline_static void
clr_mark (CompSat * compsat, int lit)
{
  clr_bit (compsat, &compsat->mark[SGN (lit)], ABS (lit));
}

/*------------------------------------------------------------------------*/

inline_static int
get_used (CompSat * compsat, int idx)
{
  return get_bit (compsat, &compsat->used, idx);
}

/*------------------------------------------------------------------------*/

inline_static void
set_used (CompSat * compsat, int idx)
{
  set_bit (compsat, &compsat->used, idx);
}

/*------------------------------------------------------------------------*/

inline_static Stack *
get_watched (CompSat * compsat, int lit)
{
  return &compsat->variables[ABS (lit)].watched[SGN (lit)];
}

/*------------------------------------------------------------------------*/

inline_static Stack *
get_implies (CompSat * compsat, int lit)
{
  return &compsat->variables[ABS (lit)].implies[SGN (lit)];
}

/*------------------------------------------------------------------------*/

inline_static int
get_assignment (CompSat * compsat, int idx)
{
  assert (idx);
  assert (idx > -compsat->size);
  assert (idx < compsat->size);
  return compsat->assignment[idx];
}

/*------------------------------------------------------------------------*/

inline_static void
set_assignment (CompSat * compsat, int idx, int assignment)
{
  compsat->assignment[idx] = assignment;
  compsat->assignment[-idx] = -assignment;
}

/*------------------------------------------------------------------------*/

inline_static void
set_reason (CompSat * compsat, int idx, int reason)
{
  assert (idx > 0);
  compsat->variables[idx].reason = reason;
}

/*------------------------------------------------------------------------*/

inline_static int
get_reason (CompSat * compsat, int idx)
{
  assert (idx > 0);
  return compsat->variables[idx].reason;
}

/*------------------------------------------------------------------------*/

inline_static void
set_pos (CompSat * compsat, int lit, int pos)
{
  compsat->variables[ABS (lit)].pos[SGN (lit)] = pos;
}

/*------------------------------------------------------------------------*/

inline_static int
get_pos (CompSat * compsat, int lit)
{
  return compsat->variables[ABS (lit)].pos[SGN (lit)];
}

/*------------------------------------------------------------------------*/

inline_static int
get_score (CompSat * compsat, int lit)
{
  return compsat->variables[ABS (lit)].score[SGN (lit)];
}

/*------------------------------------------------------------------------*/

inline_static int
max_score (CompSat * compsat, int lit)
{
  int res, tmp;
  Variable * v;

  v = compsat->variables + ABS(lit);
  res = v->score[0];
  tmp = v->score[1];
  if (tmp > res)
    res = tmp;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
set_score (CompSat * compsat, int lit, int score)
{
  compsat->variables[ABS (lit)].score[SGN (lit)] = score;
}

/*------------------------------------------------------------------------*/

inline_static void
inc_score (CompSat * compsat, int lit)
{
  int old_score, new_score, pos;

  old_score = get_score (compsat, lit);
  new_score = old_score + 1;

  assert (new_score > 0);

  if (!old_score)
    {
      pos = get_pos (compsat, lit);
      if (compsat->zero >= pos)
	compsat->zero = pos + 1;
    }

  set_score (compsat, lit, new_score);
}

/*------------------------------------------------------------------------*/

inline_static void
inc_ref (CompSat * compsat, int lit)
{
  int * p = compsat->variables[ABS (lit)].ref + SGN (lit);
  *p += 1;
  assert (*p);
}

/*------------------------------------------------------------------------*/

inline_static void
dec_ref (CompSat * compsat, int lit)
{
  int *p;
  p = compsat->variables[ABS (lit)].ref + SGN (lit);
  assert (*p > 0);
  *p -= 1;
}

/*------------------------------------------------------------------------*/

inline_static void
set_salt (CompSat * compsat, int lit, unsigned salt)
{
  compsat->variables[ABS (lit)].salt[SGN (lit)] = salt;
}

/*------------------------------------------------------------------------*/

inline_static unsigned
get_salt (CompSat * compsat, int lit)
{
  return compsat->variables[ABS (lit)].salt[SGN (lit)];
}

/*------------------------------------------------------------------------*/

inline_static int
copy_clause (CompSat * compsat, const int *clause)
{
  const int *p;
  int res;

  res = count_stack (&compsat->literals);

  for (p = clause; *p; p++)
    push_stack (compsat, &compsat->literals, *p);

  push_stack (compsat, &compsat->literals, 0);

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
unmark_visited (CompSat * compsat)
{
  int *p;

  for (p = compsat->visited.start; p < compsat->visited.top; p++)
    clr_mark (compsat, *p);

  reset_stack (compsat, &compsat->visited, 0);
}

/*------------------------------------------------------------------------*/

inline_static int
simplify_clause (CompSat * compsat, const int *clause)
{
  int trivial, lit, res;
  const int *p;

  res = count_stack (&compsat->literals);
  trivial = 0;
  p = clause;

  for (;;)
    {
      lit = *p;
      if (!lit)
	break;

      if ((trivial = get_mark (compsat, -lit)))
	break;

      p++;

      if (get_mark (compsat, lit))
	continue;

      set_mark (compsat, lit);
      push_stack (compsat, &compsat->visited, lit);

      push_stack (compsat, &compsat->literals, lit);
    }

  if (trivial)
    {
      reset_stack (compsat, &compsat->literals, res);
      res = 0;
    }
  else
    push_stack (compsat, &compsat->literals, 0);

  unmark_visited (compsat);

  return res;
}

/*------------------------------------------------------------------------*/
#if defined(COMPSAT_CHECK)
/*------------------------------------------------------------------------*/

inline_static void
enlarge_stack2 (CompSat * compsat, Stack ** stack, int old_size, int new_size)
{
  int i, new_bytes, old_bytes;

  old_bytes = old_size * sizeof (Stack);
  new_bytes = new_size * sizeof (Stack);

  for (i = 0; i <= 1; i++)
    stack[i] = resize (compsat, stack[i], old_bytes, new_bytes);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
enlarge_aux (CompSat * compsat)
{
  size_t old_bytes, new_bytes;
  int old_size, new_size;

  old_size = compsat->size;
  new_size = old_size ? 2 * old_size : 1;

  old_bytes = old_size * sizeof (compsat->variables[0]);
  new_bytes = new_size * sizeof (compsat->variables[0]);

  compsat->variables = (Variable *)
    resize (compsat, compsat->variables, old_bytes, new_bytes);

  old_bytes = 2 * old_size * sizeof (compsat->assignment[0]);
  new_bytes = 2 * new_size * sizeof (compsat->assignment[0]);

  compsat->assignment -= old_size;
  compsat->assignment = (signed char *)
    resize (compsat, compsat->assignment, old_bytes, new_bytes);
  compsat->assignment += new_size;

#ifdef COMPSAT_CHECK
  enlarge_stack2 (compsat, compsat->checker.occurence, old_size, new_size);
#endif
  compsat->size = new_size;
}

/*------------------------------------------------------------------------*/

inline_static void
enlarge (CompSat * compsat, int idx)
{
  while (idx >= compsat->size)
    enlarge_aux (compsat);

  if (idx > compsat->max)
    {
      compsat->max = idx;
      assert (idx < MAX);
    }
}

/*------------------------------------------------------------------------*/

inline_static void
insert_variables (CompSat * compsat, const int *clause)
{
  const int *p;
  int idx, max;

  max = -1;

  for (p = clause; *p; p++)
    {
      if ((idx = ABS (*p)) > max)
	max = idx;

      if (!get_used (compsat, idx))
	{
	  compsat->stats.variables++;
	  set_used (compsat, idx);
	}
    }

  enlarge (compsat, max);
}

/*------------------------------------------------------------------------*/

inline_static void
order_clause (CompSat * compsat, int clause_idx)
{
  CursorStack (&compsat->literals, it, clause_idx, 1);
  void *saved, *next, *watched, *start;
  int idx, level, max_level, count;

  save_stack_cursor (it, next);

  for (count = 0; count < WATCHED; count++)
    {
      saved = start = next;

      if (!(idx = ABS (next_stack (it))))
	break;

      save_stack_cursor (it, next);

      max_level = -1;
      watched = 0;

      do
	{
	  level = get_level (compsat, idx);
	  if (level > max_level)
	    {
	      watched = saved;
	      max_level = level;
	    }

	  save_stack_cursor (it, saved);
	}
      while ((idx = ABS (next_stack (it))));

      assert (watched);
      swap_stack (&compsat->literals, it, start, watched);
      restore_stack_cursor (it, next);
    }
}

/*------------------------------------------------------------------------*/
#if defined(COMPSAT_CHECK)
/*------------------------------------------------------------------------*/

inline_static void
release_stack2 (CompSat * compsat, Stack ** stack, int size)
{
  int i, j;

  for (i = 0; i <= 1; i++)
    {
      for (j = 0; j < size; j++)
	release_stack (compsat, stack[i] + j);

      delete (compsat, stack[i], size * sizeof (Stack));
    }
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/

inline_static int
is_connected (CompSat * compsat, int literal_idx)
{
  Stack *watched;
  int lit, tmp, res;

  lit = get_stack (&compsat->literals, literal_idx);
  watched = get_watched (compsat, lit);

  {
    IteratorStack (watched, it, 1);
    res = 0;
    while (!res && !done_stack (it))
      {
	tmp = next_stack (it);
	if (tmp == literal_idx)
	  res = 1;
      }
  }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
check_clause (CompSat * compsat, int clause_idx)
{
  CursorStack (&compsat->literals, it, clause_idx - 1, 1);
  int i, tmp;

  tmp = next_stack (it);
  assert (!tmp);

  for (i = 0; i < WATCHED && next_stack (it); i++)
    assert (is_connected (compsat, clause_idx + i));
}

/*------------------------------------------------------------------------*/

inline_static void
check (CompSat * compsat)
{
  IteratorStack (&compsat->literals, literals_it, 1);
  Stack *watched;
  int lit, i, s, tmp;

  assert (compsat->learned <= compsat->iteration);
  assert (!get_stack (&compsat->literals, compsat->learned - 1));

  tmp = next_stack (literals_it);
  assert (!tmp);

  for (i = 1; !done_stack (literals_it); i++)
    {
      check_clause (compsat, i);
      while (next_stack (literals_it))
	i++;
    }

  for (i = 1; i <= compsat->max; i++)
    for (s = -1; s <= 1; s += 2)
      {
	lit = i * s;
	watched = get_watched (compsat, lit);

	{
	  IteratorStack (watched, watched_it, 1);
	  while (!done_stack (watched_it))
	    {
	      tmp = next_stack (watched_it);
	      assert (get_stack (&compsat->literals, tmp) == lit);
	    }
	}
      }
}

/*------------------------------------------------------------------------*/

inline_static void
init_checker (Checker * checker, CompSat * compsat)
{
  checker->compsat = compsat;
  push_stack (checker->compsat, &checker->literals, 0);
}

/*------------------------------------------------------------------------*/

inline_static void
reset_checker (Checker * checker)
{
  reset_stack (checker->compsat, &checker->stack, 0);
}

/*------------------------------------------------------------------------*/

inline_static void
release_checker (Checker * checker)
{
  release_stack (checker->compsat, &checker->literals);
  release_stack (checker->compsat, &checker->units);
  release_stack (checker->compsat, &checker->stack);
  release_stack2 (checker->compsat, checker->occurence,
		  checker->compsat->size);
}

/*------------------------------------------------------------------------*/

inline_static Stack *
occurence (Checker * checker, int lit)
{
  return checker->occurence[SGN (lit)] + ABS (lit);
}

/*------------------------------------------------------------------------*/

inline_static void
add_checker (Checker * checker, int clause_idx)
{
  CursorStack (&checker->compsat->literals, it, clause_idx - 1, 1);
  Stack *stack;
  int i, lit;

  i = count_stack (&checker->literals);

  lit = next_stack (it);
  assert (!lit);

  while ((lit = next_stack (it)))
    {
      push_stack (checker->compsat, &checker->literals, lit);
      stack = occurence (checker, lit);
      push_stack (checker->compsat, stack, i++);
    }

  push_stack (checker->compsat, &checker->literals, 0);
}

/*------------------------------------------------------------------------*/

inline_static int
assignment_checker (Checker * checker, int lit)
{
  int res;

  res = checker->assignment[ABS (lit)];
  if (SGN (lit))
    res = -res;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
assign_checker (Checker * checker, int lit)
{
  int val;

  val = SGN (lit) ? FALSE : TRUE;
  checker->assignment[ABS (lit)] = val;
}

/*------------------------------------------------------------------------*/

inline_static void
check_is_implied (Checker * checker)
{
  int occidx, conflict, lit, found, assignment, tmp, dir, steps;
  CompSat *compsat;
  size_t bytes;
  Stack *occ;

  compsat = checker->compsat;
  bytes = compsat->max + 1;
  bytes *= sizeof (checker->assignment[0]);
  checker->assignment = new (compsat, bytes);

  {
    IteratorStack (&checker->units, unitit, 1);
    while (!done_stack (unitit))
      push_stack (compsat, &checker->stack, next_stack (unitit));
  }

  {
    IteratorStack (&compsat->clause, clauseit, 1);
    while (!done_stack (clauseit))
      push_stack (compsat, &checker->stack, -next_stack (clauseit));
  }

  conflict = 0;
  steps = 0;

  while (!conflict && !is_empty_stack (&checker->stack))
    {
      steps++;

      lit = pop_stack (compsat, &checker->stack);
      tmp = assignment_checker (checker, lit);

      if ((conflict = (tmp == FALSE)))
	break;

      if (tmp == TRUE)
	continue;

      assign_checker (checker, lit);
      occ = occurence (checker, -lit);

      {
	IteratorStack (occ, occit, 1);
	while (!conflict && !done_stack (occit))
	  {
	    occidx = next_stack (occit);
	    assignment = 0;
	    found = 0;

	    for (dir = -1; dir <= 1; dir += 2)
	      {
		CursorStack (&checker->literals, litit, occidx, dir);
		while (found < 2 && (lit = next_stack (litit)))
		  {
		    tmp = assignment_checker (checker, lit);
		    if (tmp == TRUE)
		      found = 2;
		    else if (tmp != FALSE)
		      {
			assert (!tmp);
			assignment = lit;
			found++;
		      }
		  }
		occidx++;
	      }

	    if (!found)
	      conflict = 1;
	    else if (found == 1)
	      push_stack (compsat, &checker->stack, assignment);
	  }
      }
    }

#ifdef COMPSAT_LOG
  log (compsat, 3, "%d CHECK %d\n", compsat->level, steps);
#endif
  assert (conflict);
  delete (compsat, checker->assignment, bytes);
  checker->assignment = 0;
  reset_checker (checker);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_LOG
/*------------------------------------------------------------------------*/

inline_static void
log_connect_with_lit (CompSat * compsat, int level, const char * msg, 
                      int lit, int pos)
{
  log (compsat, level, "%d %s %d AT %d IN CLAUSE [%d,..]\n",
       compsat->level, msg,
       lit,
       pos,
       get_stack (&compsat->literals, pos - 1) ? pos - 1: pos);
}

/*------------------------------------------------------------------------*/

inline_static void
log_connect (CompSat * compsat, int level, const char * msg, int pos)
{
  log_connect_with_lit (compsat, level, msg, 
                        get_stack (&compsat->literals, pos), pos);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
connect_literal (CompSat * compsat, int pos)
{
  int lit;

#ifdef COMPSAT_LOG
  assert (pos > 0);
  log_connect (compsat, 6, "CONNECT", pos);
#endif
  lit = get_stack (&compsat->literals, pos);
  push_stack (compsat, get_watched (compsat, lit), pos);
}

/*------------------------------------------------------------------------*/
#if defined(COMPSAT_LOG) || defined(COMPSAT_CHECK)
/*------------------------------------------------------------------------*/

void
compsat_dump_clause (CompSat * compsat, FILE * file, int clause_idx)
{
  Stack *literals = &compsat->literals;
  int i, j;

  /* The following probably is a bad style of defensive programming.
   * However it makes it easy to call 'compsat_dump_clause' from a debugger,
   * since 'stdout' may not be accessible as a symbolic name.
   */
  if (!file)
    file = stdout;

  while (get_stack (literals, clause_idx - 1))
    clause_idx--;

  for (i = 0; get_stack (literals, clause_idx + i); i++)
    ;

  fprintf (file, "CLAUSE [%d,%d]", clause_idx, clause_idx + i - 1);

  for (j = clause_idx; j < clause_idx + i; j++)
    fprintf (file, " %d", get_stack (literals, j));

  fprintf (file, "\n");
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_LOG
/*------------------------------------------------------------------------*/

inline_static void
log_clause (CompSat * compsat, int level, int clause_idx)
{
  FILE *file;

  if (compsat->verbosity >= level)
    {
      file = log_file (compsat);
      fprintf (file, COMPSAT_PREFIX "%d ", compsat->level);
      compsat_dump_clause (compsat, file, clause_idx);
    }
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
size_literals (CompSat * compsat, int clause_idx)
{
  CursorStack (&compsat->literals, it, clause_idx, 1);
  int res;

  for (res = 0; next_stack (it); res++)
    ;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
register_literals (CompSat * compsat, int clause_idx, int learning)
{
  CursorStack (&compsat->literals, it, clause_idx, 1);
  int lit, i, size;

  size = size_literals (compsat, clause_idx);

  if (size > 2 - learning)
    {
      for (i = 0; !learning && (lit = next_stack (it)); i++)
	inc_ref (compsat, lit);

      compsat->stats.literals += size;
      compsat->stats.clauses++;
    }
  else if (size == 2)
    {
      compsat->stats.implications++;
    }
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_LOG
/*------------------------------------------------------------------------*/

inline_static void
log_implies (CompSat * compsat, int level, int precondition, int conclusion)
{
  log (compsat, level, "%d IMPLIES %d %d\n",
       compsat->level, precondition, conclusion);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
set_satisfied (CompSat * compsat, int pos)
{
  assert (!get_stack (&compsat->literals, pos - 1));
  compsat->satisfied = pos;
}

/*------------------------------------------------------------------------*/

inline_static void
reset_satisfied (CompSat * compsat)
{
  set_satisfied (compsat, count_stack (&compsat->literals));
}

/*------------------------------------------------------------------------*/

inline_static void
connect_clause (CompSat * compsat, int clause_idx, int learning)
{
  int i, first_lit, second_lit, count;

  if (!(first_lit = get_stack (&compsat->literals, clause_idx)))
    {
#ifdef COMPSAT_LOG
      log (compsat, 3, "%d EMPTY\n", compsat->level);
#endif
      compsat->contains_empty_clause = 1;
      reset_stack (compsat, &compsat->literals, clause_idx);
    }
  else if (!(second_lit = get_stack (&compsat->literals, clause_idx + 1)))
    {
#ifdef COMPSAT_LOG
      log (compsat, 3, "%d UNIT %d\n", compsat->level, first_lit);
#endif
      push_stack (compsat, &compsat->units, first_lit);
      reset_stack (compsat, &compsat->literals, clause_idx);
#ifdef COMPSAT_CHECK
      if (compsat->checker.enabled)
	push_stack (compsat, &compsat->checker.units, first_lit);
#endif
    }
  else if (!learning && !get_stack (&compsat->literals, clause_idx + 2))
    {
#ifdef COMPSAT_LOG
      log_implies (compsat, 3, -first_lit, second_lit);
      log_implies (compsat, 3, -second_lit, first_lit);
#endif
      push_stack (compsat, get_implies (compsat, -first_lit), second_lit);
      push_stack (compsat, get_implies (compsat, -second_lit), first_lit);
#ifdef COMPSAT_CHECK
      add_checker (&compsat->checker, clause_idx);
#endif
      reset_stack (compsat, &compsat->literals, clause_idx);
    }
  else
    {
      CursorStack (&compsat->literals, it, clause_idx, 1);
#ifdef COMPSAT_LOG
      log_clause (compsat, 3, clause_idx);
#endif
      for (i = 0; i < WATCHED && next_stack (it); i++)
	connect_literal (compsat, i + clause_idx);

      count = count_stack (&compsat->literals);
      set_satisfied (compsat, count);
      if (!learning)
	{
	  compsat->learned = count;
	  compsat->iteration = count;
	}
#ifdef COMPSAT_CHECK
      if (compsat->checker.enabled)
	{
	  check_clause (compsat, clause_idx);
	  add_checker (&compsat->checker, clause_idx);
	}
#endif
    }
}

/*------------------------------------------------------------------------*/

inline_static int
add (CompSat * compsat)
{
  int clause_idx;

  push_stack (compsat, &compsat->clause, 0);
  clause_idx = copy_clause (compsat, compsat->clause.start);
  order_clause (compsat, clause_idx);
  register_literals (compsat, clause_idx, 1);
  connect_clause (compsat, clause_idx, 1);
  reset_stack (compsat, &compsat->clause, 0);

  return clause_idx;
}

/*------------------------------------------------------------------------*/

inline_static int
is_assigned (CompSat * compsat, int lit)
{
  return get_assignment (compsat, ABS (lit)) != 0;
}

/*------------------------------------------------------------------------*/

inline_static void
assign (CompSat * compsat, int lit, int reason)
{
  int idx, sign;

  compsat->stats.assignments++;
  if (!compsat->level)
    compsat->stats.toplevel++;

  push_stack (compsat, &compsat->trail, lit);
  assert (!get_assignment (compsat, lit));
  idx = ABS (lit);
  sign = SGN (lit);
  set_assignment (compsat, idx, sign ? FALSE : TRUE);
  set_level (compsat, idx, compsat->level);
  set_reason (compsat, idx, reason);
#ifdef COMPSAT_LOG
  log_reason_with_lit (compsat, 4, "ASSIGN", lit, reason);
#endif

}

/*------------------------------------------------------------------------*/

inline_static int
more_unprocessed_assignments (CompSat * compsat)
{
  int count;

  count = count_stack (&compsat->trail);

  if (compsat->next_implies < count)
    return 1;

  if (compsat->next_watched < count)
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

inline_static int
dequeue_implies (CompSat * compsat)
{
  int res, * p;

  assert (compsat->next_implies < count_stack (&compsat->trail));
  p = compsat->trail.start + compsat->next_implies++;
  res = *p;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
imply (CompSat * compsat)
{
  int lit = dequeue_implies (compsat), tmp, other;
  Stack * implies = get_implies (compsat, lit);
  int new_reason, reason, conflict;
  IteratorStack (implies, it, 1);

  if (is_empty_stack (implies))
    return 0;

#ifdef COMPSAT_LOG
  log (compsat, 6, "%d IMPLY %d\n", compsat->level, lit);
#endif
  reason = get_reason (compsat, ABS(lit));
  assert (reason);

  while (!done_stack (it))
    {
      other = next_stack (it);
#ifdef COMPSAT_LOG
      log (compsat, 6, "%d IMPLIED %d\n", compsat->level, other);
#endif
      tmp = get_assignment (compsat, other);

      if (tmp == TRUE)
	continue;

      if (tmp == FALSE)
	{
	  compsat->conflict = reason;
	  conflict = get_reason (compsat, ABS (other));
	  assert (conflict);
	  return conflict;
	}

      assert (!tmp);
      if (is_clause_reason (reason))
	new_reason = reason;
      else if (is_unit_reason (reason))
	new_reason = unit2reason (other);
      else 
	{
	  assert (is_decision_reason (reason));
	  new_reason = decision2reason (other);
	}

      assign (compsat, other, new_reason);
    }

  return 0;
}

/*------------------------------------------------------------------------*/

inline_static int
dequeue_watched (CompSat * compsat)
{
  int res, * p;

  assert (compsat->next_watched < count_stack (&compsat->trail));
  p = compsat->trail.start + compsat->next_watched++;
  res = *p;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
propagate (CompSat * compsat, int pos, int *remove_ptr)
{
  int lit, tmp, assignment, old_lit, other, other_was_false;
  CursorStack (&compsat->literals, it, pos - 1, 1);
  void *old, *last;

  assert (WATCHED == 2);
  *remove_ptr = 0;

  other = next_stack (it);
  save_stack_cursor (it, old);
  old_lit = next_stack (it);
  assert (get_assignment (compsat, old_lit) == FALSE);

#ifdef COMPSAT_LOG
  log (compsat, 6, "%d PROPAGATE %d AT %d IN CLAUSE [%d,..]\n",
       compsat->level, old_lit, pos,
       get_stack (&compsat->literals, pos - 1) ? pos - 1 : pos);
#endif
  if (other)
    {
      assert (!get_stack (&compsat->literals, pos - 2));
      tmp = get_assignment (compsat, other);
      assignment = pos - 1;
    }
  else if ((other = next_stack (it)))
    {
      tmp = get_assignment (compsat, other);
      assignment = pos + 1;
    }
  else
    return pos;

  if (tmp == TRUE)
    return 0;

  /* Since we assign the literals eagerly the other watched literal may be
   * assigned already, but not propagated yet.  Therefore we still need
   * to traverse the whole clause before potentially concluding
   * unsatisfiability of this clause.
   */
  other_was_false = (tmp == FALSE);

  /* Now search through the non watched literals of the clause for a non
   * FALSE assigned literal.  If one is found, then disconnect the 'old'
   * literal and connect the new literal.  In addition swap the position of
   * the non false literal with the start literal.
   */
  save_stack_cursor (it, last);
  while ((lit = next_stack (it)))
    {
      tmp = get_assignment (compsat, lit);

      if (tmp != FALSE)
	{
#ifdef COMPSAT_LOG
	  log_connect (compsat, 6, "DISCONNECT", pos);
#endif
	  swap_stack (&compsat->literals, it, old, last);
	  *remove_ptr = 1;
	  connect_literal (compsat, pos);

	  return 0;
	}

      save_stack_cursor (it, last);
    }

#ifndef NDEBUG
  tmp = get_assignment (compsat, other);
  assert (tmp != TRUE);
  assert (other_was_false == (tmp == FALSE));
#endif
  if (other_was_false)
    {
      assert (assignment > 0);
      return assignment;
    }

  assign (compsat, other, assignment);

  return 0;
}

/*------------------------------------------------------------------------*/

inline_static int
watch (CompSat * compsat)
{
  int tmp, lit, conflict, remove, *p, *last;
  Stack *stack;

  lit = dequeue_watched (compsat);
  tmp = get_assignment (compsat, lit);
  assert (tmp == TRUE);

  stack = get_watched (compsat, -lit);
  if (is_empty_stack (stack))
    return 0;

  p = stack->start;
  last = stack->top;
  conflict = 0;

  while (!conflict && p < last)
    {
      conflict = propagate (compsat, *p, &remove);

      if (remove)
	*p = *--last;
      else
	p++;
    }

  reset_stack (compsat, stack, last - stack->start);

  if (conflict)
    compsat->stats.conflicts++;

  return conflict;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_LOG
/*------------------------------------------------------------------------*/

inline_static void
log_conflict (CompSat * compsat, int level, int reason)
{
  log_reason (compsat, level, "CONFLICT ON", reason);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
deduce (CompSat * compsat)
{
  int conflict, count;

  conflict = 0;
  while (!conflict)
    {
      count = count_stack (&compsat->trail);

      if (compsat->next_implies < count)
	{
	  conflict = imply (compsat);
	  continue;
	}

      if (compsat->next_watched >= count)
	break;

      conflict = watch (compsat);
    }
#ifdef COMPSAT_LOG
  if (conflict)
    {
      log_conflict (compsat, 3, conflict);
      if (compsat->conflict && 
	  reason2lit (compsat, conflict) != 
	    reason2lit (compsat, compsat->conflict))
	log_conflict (compsat, 3, compsat->conflict);
    }
#endif

  return conflict;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_CHECK
/*------------------------------------------------------------------------*/

inline_static void
check_is_satisfied (CompSat * compsat)
{
  IteratorStack (&compsat->literals, it, 1);
  int clause_satisfied, lit, tmp;

  tmp = next_stack (it);
  assert (!tmp);

  while (!done_stack (it))
    {
      clause_satisfied = 0;

      while ((lit = next_stack (it)))
	{
	  if (!clause_satisfied)
	    {
	      tmp = get_assignment (compsat, lit);
	      clause_satisfied = (tmp == TRUE);
	    }
	}

      assert (clause_satisfied);
    }
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static double
avglevel (CompSat * compsat)
{
  double res;

  if (compsat->stats.decisions > 0)
    res = compsat->stats.sum_levels_at_decision / compsat->stats.decisions;
  else 
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
report_header (CompSat * c)
{
  log (c, 1, "\n");
  log (c, 1, "  seconds     iterations          conflicts        avglevel\n");
  log (c, 1,
"           MB          assignments        decisions"
#ifdef DECOMPOSE
"      components"
#endif
"\n");
  log (c, 1, "\n");
}

/*------------------------------------------------------------------------*/

inline_static void
report (CompSat * compsat)
{
  double mb;

  if (compsat->verbosity >= 1)
    {
      compsat->stats.reported++;

      if (compsat->stats.reported == 1)
	report_header (compsat);

      update_time (compsat);
      mb = compsat_bytes (compsat);
      mb /= 1024.0;
      mb /= 1024.0;

      log (compsat, 1,
	   "%7.2f %7.1f %5d %10.0f %8.0f %8.0f %7.1f"
#ifdef DECOMPOSE
	   "%6d"
#endif
	   "\n",
	   compsat->stats.time,
	   mb,
	   compsat->stats.iterations,
	   compsat->stats.assignments,
	   compsat->stats.conflicts,
	   compsat->stats.decisions,
	   avglevel (compsat)
#ifdef DECOMPOSE
	   , compsat->stats.non_trivial_components
#endif
	   );
    }
}

/*------------------------------------------------------------------------*/
#ifdef DECOMPOSE
/*------------------------------------------------------------------------*/

inline_static int
deref_component (CompSat * compsat, int idx)
{
  int res, next;

  res = idx;
  while ((next = get_component (compsat, res)))
    res = next;

  return res;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
cmp (CompSat * compsat, int a, int b)
{
  int score[2], salt[2], idx[2], sign[2], impls[2];
#ifdef DECOMPOSE
  int component[2], size[2];
#endif

  if (a == b)
    return 0;

  if (a == -b)
    {
      score[0] = get_score (compsat, a);
      score[1] = get_score (compsat, b);
      if (score[0] < score[1])
	return -1;
      if (score[0] > score[1])
	return 1;
    }
  else
    {
#ifdef DECOMPOSE
      size[0] = get_component_size (compsat, ABS(a));
      size[1] = get_component_size (compsat, ABS(b));
      if (size[0] > size[1])
	return -1;
      if (size[0] < size[1])
	return 1;

      component[0] = deref_component (compsat, ABS(a));
      component[1] = deref_component (compsat, ABS(b));
      if (component[0] > component[1])
	return -1;
      if (component[0] < component[1])
	return 1;
#endif
      score[0] = max_score (compsat, a);
      score[1] = max_score (compsat, b);
      if (score[0] < score[1])
	return -1;
      if (score[0] > score[1])
	return 1;
    }

  impls[0] = count_stack (get_implies (compsat, a));
  impls[1] = count_stack (get_implies (compsat, b));
  if (impls[0] < impls[1])
    return -1;
  if (impls[0] > impls[1])
    return 1;

  salt[0] = get_salt (compsat, a);
  salt[1] = get_salt (compsat, b);
  if (salt[0] < salt[1])
    return -1;
  if (salt[0] > salt[1])
    return 1;

  idx[0] = ABS (a);
  idx[1] = ABS (b);
  if (idx[0] > idx[1])
    return -1;
  if (idx[0] < idx[1])
    return 1;

  sign[0] = SGN (a);
  sign[1] = SGN (b);
  if (sign[0] < sign[1])
    return -1;
  if (sign[0] > sign[1])
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

inline_static int
best_phase (CompSat * compsat, int lit)
{
  return (cmp (compsat, -lit, lit) < 0) ? lit : -lit;
}

/*------------------------------------------------------------------------*/

inline_static void
new_level (CompSat * compsat, int lit)
{
  int assignment;

  assert (!is_assigned (compsat, lit));
  compsat->level++;
  push_stack (compsat, &compsat->control, count_stack (&compsat->trail));
  assert (compsat->level == count_stack (&compsat->control));

  assert (best_phase (compsat, lit) == lit);
  assignment = decision2reason (lit);
  assign (compsat, lit, assignment);
}

/*------------------------------------------------------------------------*/

inline_static int
berkmin (CompSat * compsat)
{
  int *p, *end, lit, best, tmp, csearched, lsearched, res, * literals;
#ifdef DECOMPOSE
  int component;
#endif

  best = 0;
  literals = compsat->literals.start;
  p = literals + compsat->satisfied;
  assert (compsat->learned <= compsat->iteration);
  end =  literals + compsat->iteration;
  csearched = lsearched = 0;

  if (p <= end)
    return 0;

  p--;
  res = 0;

  while (!res && p >= end)
    {
      lit = *--p;
      if (lit)
	{
	  lsearched++;
	  tmp = get_assignment (compsat, lit);

	  if (tmp == FALSE)
	    continue;

	  if (tmp == TRUE)
	    {
	      csearched++;

	      while (*--p)
		lsearched++;

	      set_satisfied (compsat, (p + 1) - literals);
	      best = 0;

	      continue;
	    }
#ifdef DECOMPOSE
	  if (compsat->active)
	    {
	      component = deref_component (compsat, ABS (lit));
	      if (component != compsat->active)
		continue;
	    }
#endif
	  lit = best_phase (compsat, lit);

	  assert (!tmp);
	  if (!best || cmp (compsat, lit, best) > 0)
	    best = lit;

	  assert (best == best_phase (compsat, best));
	}
      else if (best)
	{
#ifdef COMPSAT_LOG
	  log (compsat, 3, "%d BERKMIN %d\n", compsat->level, best);
#endif
	  assert (best);
	  res = best;
	}
      else
	set_satisfied (compsat, (p + 2) - literals);
    }

  compsat->stats.csearched += csearched;
  compsat->stats.lsearched += lsearched;
#ifdef COMPSAT_LOG
  log (compsat, 3, "%d SEARCHED %d CLAUSES AND %d LITERALS\n",
       compsat->level, csearched, lsearched);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
vsids (CompSat * compsat)
{
  int *p, * start, *end, lit, searched, res;

  start = compsat->order.start + compsat->unassigned;
  end = compsat->order.top;
  res = 0;

  for (p = start; !res && p < end; p++)
    {
      lit = *p;
      if (is_assigned (compsat, lit))
	continue;

      lit = best_phase (compsat, lit);
      searched = p - start;
      compsat->stats.osearched += searched;
#ifdef COMPSAT_LOG
      log (compsat, 3, "%d VSIDS %d\n", compsat->level, lit);
      log (compsat, 3, "%d OSEARCHED %d\n", compsat->level, searched);
#endif
      compsat->unassigned = p - compsat->order.start + 1;
      assert (lit);
      res = lit;
    }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
assign_units (CompSat * compsat)
{
  IteratorStack (&compsat->units, it, 1);
  int lit, tmp, assignment;

  compsat->stats.units += count_stack (&compsat->units);

  while (!done_stack (it))
    {
      lit = next_stack (it);
      tmp = get_assignment (compsat, lit);

      if (tmp == TRUE)
	{
#ifdef COMPSAT_LOG
	  log (compsat, 3, "unit %d occurs twice\n", lit);
#endif
	  continue;
	}

      if (tmp == FALSE)
	{
#ifdef COMPSAT_LOG
	  log (compsat, 3, "conflicting unit %d\n", lit);
#endif
	  return 0;
	}

      assignment = unit2reason (lit);
      assign (compsat, lit, assignment);
    }

  return 1;
}

/*------------------------------------------------------------------------*/

inline_static void
swap (int *a, int *b)
{
  int tmp;

  tmp = *a;
  *a = *b;
  *b = tmp;
}

/*------------------------------------------------------------------------*/

inline_static void
cmpswap (CompSat * compsat, int *a, int *b)
{
  if (cmp (compsat, *a, *b) < 0)
    swap (a, b);
}

/*------------------------------------------------------------------------*/

inline_static int
partition (CompSat * compsat, int *a, int l, int r)
{
  int lit, i, j;

  i = l - 1;
  j = r;
  lit = a[r];

  for (;;)
    {
      while (cmp (compsat, a[++i], lit) > 0)
	;

      while (cmp (compsat, lit, a[--j]) > 0)
	if (j == l)
	  break;

      if (i >= j)
	break;

      swap (a + i, a + j);
    }

  swap (a + i, a + r);

  return i;
}

/*------------------------------------------------------------------------*/

inline_static void
insertion (CompSat * compsat, int *a, int l, int r)
{
  int lit, i, j;

  for (i = r; i > l; i--)
    cmpswap (compsat, a + i - 1, a + i);

  for (i = l + 2; i <= r; i++)
    {
      j = i;
      lit = a[i];
      while (cmp (compsat, lit, a[j - 1]) > 0)
	{
	  a[j] = a[j - 1];
	  j--;
	}
      a[j] = lit;
    }
}

/*------------------------------------------------------------------------*/

inline_static void
quicksort (CompSat * compsat, int *a, int l, int r)
{
  int i, m, ll, rr;
  Stack *stack;

  if (r - l <= ISORT_LIMIT)
    return;

  stack = &compsat->stack;
  reset_stack (compsat, stack, 0);

  for (;;)
    {
      m = (l + r) / 2;
      swap (a + m, a + r - 1);
      cmpswap (compsat, a + l, a + r - 1);
      cmpswap (compsat, a + l, a + r);
      cmpswap (compsat, a + r - 1, a + r);
      i = partition (compsat, a, l + 1, r - 1);

      if (i - l < r - i)
	{
	  ll = i + 1;
	  rr = r;
	  r = i - 1;
	}
      else
	{
	  ll = l;
	  rr = i - 1;
	  l = i + 1;
	}

      if (r - l > ISORT_LIMIT)
	{
	  assert (rr - ll > ISORT_LIMIT);
	  push_stack (compsat, stack, ll);
	  push_stack (compsat, stack, rr);
	}
      else if (rr - ll > ISORT_LIMIT)
	{
	  l = ll;
	  r = rr;
	}
      else if (count_stack (stack))
	{
	  r = pop_stack (compsat, stack);
	  l = pop_stack (compsat, stack);
	}
      else
	break;
    }
}

/*------------------------------------------------------------------------*/

inline_static void
hypersort (CompSat * compsat, int *a, int n)
{
  quicksort (compsat, a, 0, n - 1);
  insertion (compsat, a, 0, n - 1);
}

/*------------------------------------------------------------------------*/

inline_static void
adjust (CompSat * compsat)
{
  int i, lit, last, last_non_zero_pos, first_unassigned;

  last = compsat->zero - 1;
  last_non_zero_pos = -1;
  first_unassigned = -1;

  for (i = 0; i < compsat->zero; i++)
    {
      lit = get_stack (&compsat->order, i);
      if (first_unassigned < 0 && !is_assigned (compsat, lit))
	first_unassigned = i;
      if (get_score (compsat, lit))
	last_non_zero_pos = i;
      set_pos (compsat, lit, i);
    }

  compsat->zero = last_non_zero_pos + 1;
  compsat->unassigned = (first_unassigned < 0) ? 0 : first_unassigned;
}

/*------------------------------------------------------------------------*/

inline_static void
sort (CompSat * compsat)
{
#ifdef COMPSAT_LOG
  log (compsat, 2, "%d SORT %d\n", compsat->level, compsat->zero);
#endif
  hypersort (compsat, compsat->order.start, compsat->zero);
  adjust (compsat);
}

/*------------------------------------------------------------------------*/
/* See 'Numerical Recipes in C'.
 */
inline_static unsigned
rng (CompSat * compsat)
{
  unsigned res, next;

  next = compsat->rng;
  res = next;
  next *= 1664525;
  next += 1013904223;
  compsat->rng = next;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
order (CompSat * compsat)
{
  int idx, lit, pos, sign;

  pos = 0;
  for (idx = 1; idx <= compsat->max; idx++)
    for (sign = 1; sign >= 0; sign--)
      {
	lit = (sign ? -1 : 1) * idx;
	set_stack (compsat, &compsat->order, pos, lit);
	set_pos (compsat, lit, pos);
	set_salt (compsat, lit, rng (compsat));
	pos++;
      }
  assert (pos == count_stack (&compsat->order));
  compsat->unassigned = 0;
  compsat->zero = pos;

  sort (compsat);
}

/*------------------------------------------------------------------------*/

inline_static void
rescore (CompSat * compsat)
{
  int pos, lit, score, zero;

#ifdef COMPSAT_LOG
  log (compsat, 2, "%d RESCORE %.0f\n",
       compsat->level, compsat->counters.rescore.finished);
#endif

  zero = -1;
  for (pos = compsat->zero - 1; pos >= 0; pos--)
    {
      lit = get_stack (&compsat->order, pos);
      score = get_score (compsat, lit);
      score /= 2;
      if (score && pos > zero)
	zero = pos + 1;
      set_score (compsat, lit, score);
    }

  compsat->zero = zero;
}

/*------------------------------------------------------------------------*/

inline_static int
expand_lit (CompSat * compsat, int lit, int paths)
{
  int idx, level, tmp, new_lit;

  idx = ABS (lit);
  level = get_level (compsat, idx);

  if (!level)
    return paths;

  if (get_mark (compsat, idx))
    return paths;

  set_mark (compsat, idx);
  push_stack (compsat, &compsat->visited, idx);

  tmp = get_assignment (compsat, lit);
  new_lit = (tmp == TRUE) ? -lit : lit;
  inc_score (compsat, new_lit);

  if (level == compsat->level)
    paths++;
  else
    {
      inc_score (compsat, lit);
      push_stack (compsat, &compsat->clause, lit);
    }

  return paths;
}

/*------------------------------------------------------------------------*/

inline_static int
expand_clause (CompSat * compsat, int reason, int paths)
{
  int lit, dir;

  assert (is_clause_reason (reason));
  assert (0 < reason);
  assert (reason < count_stack (&compsat->literals));

  for (dir = -1; dir <= 1; dir += 2)
    {
      CursorStack (&compsat->literals, it, reason + (dir == 1), dir);
      while ((lit = next_stack (it)))
	paths = expand_lit (compsat, lit, paths);
    }

  return paths;
}

/*------------------------------------------------------------------------*/

inline_static int
expand (CompSat * compsat, int reason, int paths)
{
  int lit;

  if (is_clause_reason (reason))
    return expand_clause (compsat, reason, paths);

  lit = reason2lit (compsat, reason);
  return expand_lit (compsat, lit, paths);
}

/*------------------------------------------------------------------------*/

inline_static void
analyze (CompSat * compsat, int conflict)
{
  int paths, *p, lit, idx, reason;

  assert (compsat->level > 0);

  paths = expand (compsat, conflict, 0);
  if (compsat->conflict)
    paths = expand (compsat, compsat->conflict, paths);

  p = compsat->trail.top;
  lit = 0;

  do
    {
      lit = *--p;
      idx = ABS (lit);

      if (get_mark (compsat, idx))
	{
	  paths--;
	  if (paths)
	    {
	      reason = get_reason (compsat, idx);
	      paths = expand (compsat, reason, paths);
	    }
	  else
	    push_stack (compsat, &compsat->clause, -lit);
	}
    }
  while (paths);

  unmark_visited (compsat);

#ifdef COMPSAT_LOG
  assert (lit);
  if (compsat->level >= 1 && lit != current_decision (compsat))
    log (compsat, 3, "%d UIP %d\n", compsat->level, lit);
#endif
#ifdef COMPSAT_CHECK
  if (compsat->checker.enabled)
    check_is_implied (&compsat->checker);
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
drive (CompSat * compsat, int clause_idx)
{
  CursorStack (&compsat->literals, it, clause_idx, 1);
  int lit, next_lit, res, count, level;

  lit = next_stack (it);
  assert (lit);
  next_lit = next_stack (it);

  if (next_lit)
    {
      assert (clause_idx < count_stack (&compsat->literals));

      for (count = clause_idx, res = 0; !res && lit; count++)
	{
	  level = get_level (compsat, ABS (lit));
	  if (level == compsat->level)
	    res = count;
	  else if ((lit = next_lit))
	    next_lit = next_stack (it);
	}
    }
  else
    {
      assert (clause_idx == count_stack (&compsat->literals));
      res = unit2reason (lit);
    }

  assert (res);

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
jump (CompSat * compsat, int clause_idx)
{
  CursorStack (&compsat->literals, it, clause_idx, 1);
  int tmp, res, lit;

  res = 0;

  while ((lit = next_stack (it)))
    {
      tmp = get_level (compsat, ABS (lit));
      if (tmp < compsat->level && tmp > res)
	res = tmp;
    }

#ifdef COMPSAT_LOG
  log (compsat, 3, "%d BACK%s %d\n",
       compsat->level, (res < compsat->level - 1) ? "JUMP" : "TRACK", res);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
untrail (CompSat * compsat, int level)
{
  int *start, *last, *p, frame_start, pos_in_order, lit;

  frame_start = get_stack (&compsat->control, level);
  start = compsat->trail.start + frame_start;
  last = compsat->trail.top - 1;

  for (p = last; p >= start; p--)
    {
      lit = *p;
      if (!get_level (compsat, ABS(lit)))
	{
#ifdef COMPSAT_LOG
	  log (compsat, 4, "%d LITERAL %d HAS BEEN SAVED\n",
	       compsat->level, lit);
#endif
	  continue;
	}

#ifdef COMPSAT_LOG
      log (compsat, 4, "%d UNASSIGN %d\n", compsat->level, lit);
#endif
      set_assignment (compsat, ABS (lit), 0);
      pos_in_order = get_pos (compsat, lit);
      if (pos_in_order < compsat->unassigned)
	compsat->unassigned = pos_in_order;
    }

  compsat->next_implies = frame_start;
  compsat->next_watched = frame_start;
  compsat->level = level;
  reset_stack (compsat, &compsat->control, level);
  reset_stack (compsat, &compsat->trail, frame_start);
  compsat->conflict = 0;
  reset_satisfied (compsat);
#ifdef DECOMPOSE
  compsat->active = 0;
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
assign_pure_literals (CompSat * compsat)
{
  int idx, lit, tmp, num_pure_literals, num_zombies, assignment, conflict;
  Variable * v;

  num_pure_literals = num_zombies = 0;
  for (idx = 1; idx <= compsat->max; idx++)
    {
      v = compsat->variables + idx;
      if (!v->ref[0] && !count_stack (v->implies + 1))
	lit = -idx;
      else if (!v->ref[1] && !count_stack (v->implies + 0))
	lit = idx;
      else 
	continue;

      tmp = get_assignment (compsat, lit);
      if (tmp)
	continue;

      if (!v->ref[0] && !v->ref[1])
	num_zombies++;
      else
	num_pure_literals++;

#ifdef COMPSAT_LOG
      log (compsat, 3, "%d UNATE %d\n", compsat->level, lit);
#endif
      assignment = unit2reason (lit);
      assign (compsat, lit, assignment);
    }

  compsat->stats.pure_literals += num_pure_literals;
  compsat->stats.zombies += num_zombies;
#ifdef COMPSAT_LOG
  if (num_pure_literals)
    log (compsat, 2, "assigned %d pure literals\n", num_pure_literals);
  if (num_zombies)
    log (compsat, 2, "assigned %d zombies\n", num_zombies);
#endif
  conflict = deduce (compsat);
  assert (!conflict);

  return num_pure_literals;
}

/*------------------------------------------------------------------------*/

inline_static void
ref2score (CompSat * compsat)
{
  Variable * v;
  int i, idx;

  for (idx = 1; idx <= compsat->max; idx++)
    {
      v = compsat->variables + idx;
      for (i = 0; i <= 1; i++)
	v->score[i] = v->ref[i];
    }
}

/*------------------------------------------------------------------------*/
#ifdef DECOMPOSE
/*------------------------------------------------------------------------*/

inline_static void
compactify_component_chain (CompSat * compsat, int idx, int repr)
{
  int next, prev;

  prev = idx;
  while (prev && prev != repr)
    {
      next = get_component (compsat, prev);
      set_component (compsat, prev, repr);
      prev = next;
    }
}

/*------------------------------------------------------------------------*/

inline_static void
decompose (CompSat * compsat)
{
  int idx, other, lit, sign, repr, count, component;
  int * p, * q, * literals, * eol, size;
  double start_time = compsat_get_time ();
  Stack * implies;

  compsat->stats.decompositions++;
#ifdef COMPSAT_LOG
  log (compsat, 2, "DECOMPOSITION %d START\n", compsat->stats.decompositions);
#endif
  for (idx = 1; idx <= compsat->max; idx++)
    {
      set_component (compsat, idx, 0);
      set_component_size (compsat, idx, 0);
    }

  literals = compsat->literals.start;
  eol = literals + compsat->learned;

  for (p = literals + 1; p < eol; p = q + 1)
    {
      repr = INT_MAX;
      for  (q = p; (lit = *q); q++)
	{
	  idx = ABS (lit);
	  component = deref_component (compsat, idx);
	  if (component < repr)
	    repr = component;
	}

      for  (q = p; (lit = *q); q++)
	{
	  idx = ABS (lit);
	  compactify_component_chain (compsat, idx, repr);
	}
    }

  for (idx = 1; idx <= compsat->max; idx++)
    {
      repr = deref_component (compsat, idx);

      for (sign = -1; sign <= 1; sign += 2)
	{
	  lit = sign * idx;
	  implies = get_implies (compsat, lit);

	  for (p = implies->start; p < implies->top; p++)
	    {
	      other = ABS (*p);
	      component = deref_component (compsat, other);
	      if (component < repr)
		repr = component;
	    }
	}

      compactify_component_chain (compsat, idx, repr);

      for (sign = -1; sign <= 1; sign += 2)
	{
	  lit = sign * idx;
	  implies = get_implies (compsat, lit);

	  for (p = implies->start; p < implies->top; p++)
	    {
	      other = ABS (*p);
	      compactify_component_chain (compsat, other, repr);
	    }
	}
    }

  count = 0;
  for (idx = 1; idx <= compsat->max; idx++)
    {
      if (!(component = get_component (compsat, idx)))
	{
	  component = idx;
	  count++;
	}
      else
	{
	  component = deref_component (compsat, idx);
	  compactify_component_chain (compsat, idx, component);
	}
#ifdef COMPSAT_LOG
      log (compsat, 6,
	   "DECOMPOSITION %d VARIABLE %d MEMBER OF COMPONENT %d\n",
	   compsat->stats.decompositions, idx, component);
#endif
    }
#ifdef COMPSAT_LOG
  log (compsat, 2, "DECOMPOSITION %d FOUND %d COMPONENTS\n",
       compsat->stats.decompositions, count);
#endif

  for (idx = 1; idx <= compsat->max; idx++)
    {
      component = deref_component (compsat, idx);
      inc_component_size (compsat, component);
    }

  for (idx = 1; idx <= compsat->max; idx++)
    {
      component = deref_component (compsat, idx);
      size = get_component_size (compsat, component);
      set_component_size (compsat, idx, size);
    }

  assert (count >= compsat->stats.components);
  compsat->stats.components = count;

  compsat->stats.non_trivial_components = 0;
  for (idx = 1; idx <= compsat->max; idx++)
    {
      if (idx != deref_component (compsat, idx))
	continue;

      if (get_component_size (compsat, idx) <= 1)
	continue;

      compsat->stats.non_trivial_components++;
    }
#ifdef COMPSAT_LOG
  log (compsat, 2, "DECOMPOSITION %d FOUND %d NON TRIVIAL COMPONENTS\n",
       compsat->stats.decompositions, compsat->stats.non_trivial_components);
#endif
  if (compsat->stats.max_non_trivial_components < 
      compsat->stats.non_trivial_components)
	compsat->stats.max_non_trivial_components = 
	compsat->stats.non_trivial_components;

  compsat->stats.decomposition_time += compsat_get_time () - start_time;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static void
gc_implies (CompSat * compsat)
{
  int collected, idx, sign, lit, * p, * q, * eoi, other, tmp, count;
  Stack * implies;

  assert (!compsat->level);
  assert (!more_unprocessed_assignments (compsat));

  collected = 0;
  for (idx = 1; idx <= compsat->max; idx++)
    {
      for (sign = -1; sign <= 1; sign += 2)
	{
	  lit = sign * idx;
	  implies = get_implies (compsat, lit);
	  tmp = get_assignment (compsat, lit);
	  if (tmp)
	    {
	      count = count_stack (implies);
	      if (!count)
		continue;

#ifdef COMPSAT_LOG
	      log (compsat, 3, 
		   "GC %d COLLECTING ALL %d IMPLICATIONS FOR %d\n", 
		   compsat->stats.gcs, count, lit);
#endif
	      collected += count;
	      release_stack (compsat, implies);
	    }
	  else
	    {
	      p = q = implies->start;
	      eoi = implies->top;
	      while (p < eoi)
		{
		  other = *p++;
		  tmp = get_assignment (compsat, other);
		  if (tmp == TRUE)
		    {
#ifdef COMPSAT_LOG
		      log (compsat, 3, "GC %d COLLECTING IMPLICATION %d %d\n", 
			   compsat->stats.gcs, lit, other);
#endif
		      collected++;
		      continue;
		    }

		  assert (tmp != FALSE);
		  *q++ = other;
		}

	      implies->top = q;
	    }
	}
    }

  compsat->stats.collected_implications += collected;
#ifdef COMPSAT_LOG
  if (collected)
    log (compsat, 2, "GC %d COLLECTED %d IMPLICATIONS\n",
	 compsat->stats.gcs, collected);
#endif
}

/*------------------------------------------------------------------------*/

inline_static int *
sticky_clause_start (CompSat * compsat)
{
  int * p, count, lit;
#if 0
  int ldrescore;

  for (ldrescore = 0; 
       (1 << ldrescore) < compsat->counters.rescore.init;
       ldrescore++)
    ;

  count = ldrescore/2 * compsat->counters.rescore.init;
#else
  count = 3 * compsat->counters.rescore.init;
#endif
  p = compsat->literals.top - 1;
  while (p > compsat->literals.start)
    {
      lit = *--p;
      if (lit)
	continue;

      count--;
      if (!count)
	break;
    }

  assert (!*p);

  return p + 1;
}

/*------------------------------------------------------------------------*/

inline_static void
gc_watched (CompSat * compsat)
{
  int lit, tmp, idx, sign, * literals, pos, score, * sticky;
  int shrunken_clauses, collected_literals, collected_clauses, count;
  int * old_learned, * new_learned, * old_iteration, * new_iteration;
  int moved_clauses, moved_literals, collect;
  int * p, * q, * eol, *new_eol, * eow;
  int num_true, num_false, size;
  Stack * watched;
#ifdef DECOMPOSE
  int component, repr;
#endif

  assert (!compsat->level);
  assert (!more_unprocessed_assignments (compsat));

  /* 1. Mark all the garbage clauses by overwriting their watched literals
   * with '0'.  For all the other clauses the watched literals are overwritten
   * with their new position.
   */
  collected_clauses = shrunken_clauses = collected_literals = 0;
  literals = compsat->literals.start;

  old_learned = literals + compsat->learned;
  sticky = sticky_clause_start (compsat);
  old_iteration = literals + compsat->iteration;
  if (old_iteration > sticky);
    sticky = old_iteration;

  eol = compsat->literals.top;
  pos = 1;

  for (p = literals + 1; p < eol; p += size + 1)
    {
#ifdef DECOMPOSE
      repr = 0;
#endif
      score = num_true = num_false = 0;
      for (q = p; (lit = *q); q++)
	{
	  tmp = get_assignment (compsat, lit);
	  if (tmp == FALSE)
	    num_false++;
	  else if (tmp == TRUE)
	    num_true++;
	  else if (!num_true)
	    {
	      score += max_score (compsat, lit);
	      idx = ABS(lit);
#ifdef DECOMPOSE
	      component = deref_component (compsat, idx);
	      if (repr)
		{
		  if (repr != component)
		    repr = INT_MIN;
		}
	      else
		repr = component;
#endif
	    }
	}
      size = q - p;

#ifdef DECOMPOSE
      if (repr == INT_MIN)
#endif
	assert (num_true || p >= old_learned);
#if 1
      if (num_true)
	collect = 1;
      else if (p < old_learned)
	collect = 0;
#ifdef DECOMPOSE
      else if (repr == INT_MIN)
	{
	  compsat->stats.cross_component_collected_learned_clauses++;
	  collect = 1;
	}
#endif
      else if (p >= sticky)
	collect = 0;
      else if ((size - num_false) <= 4)
	collect = 0;
      else if ((size - num_false) >= 1000)
	collect = 1;
      else if (score <  2 * (size - num_false))
	collect = 1;
      else
	collect = 0;
#else
      collect = num_true;
#endif

      if (collect)
	{
#ifdef COMPSAT_LOG
	  log (compsat, 3, "GC %d COLLECTING CLAUSE [%d,%d] OF SIZE %d\n", 
	       compsat->stats.gcs, p - literals, q - literals - 1, size);
#endif
	  if (p < old_learned)
	    {
	      for (q = p; (lit = *q); q++)
		dec_ref (compsat, lit);
	    }

	  p[0] = GARBAGE;
	  p[1] = GARBAGE;

	  collected_clauses++;
	  collected_literals += size;
	}
      else
	{
	  assert (!get_assignment (compsat, p[0]));
	  assert (!get_assignment (compsat, p[1]));
	  assert (size >= num_false + 2);

	  if (num_false && p < old_learned)
	    {
	      for (q = p; (lit = *q); q++)
		{
		  tmp = get_assignment (compsat, lit);
		  assert (tmp != TRUE);
		  if (tmp == FALSE)
		    dec_ref (compsat, lit);
		}
	    }

	  assert (pos > 0);
	  p[0] = pos;
	  p[1] = pos + 1;
	  pos += (size - num_false) + 1;
#ifdef COMPSAT_LOG
	  if (num_false)
	    log (compsat,  3, 
		 "GC %d SHRINKING CLAUSE [%d,%d] BY %d LITERALS\n",
		compsat->stats.gcs, 
		p - literals, q - literals - 1,
		num_false);

	  if (p[0] != p - literals || num_false)
	    log (compsat, 3, "GC %d MOVING CLAUSE [%d,%d] TO [%d,%d]\n",
		 compsat->stats.gcs,
		 p - literals, q - literals - 1,
		 p[0], p[0] + (size - num_false) - 1);
#endif
	  if (!num_false)
	    continue;

	  collected_literals += num_false;
	  shrunken_clauses++;
	}
    }
  new_eol = literals + pos;
  assert (eol - new_eol == collected_literals + collected_clauses);
#ifdef COMPSAT_LOG
  if (collected_clauses)
    log (compsat, 2, "GC %d COLLECTED %d CLAUSES\n", 
	 compsat->stats.gcs, collected_clauses);
  if (shrunken_clauses)
    log (compsat, 2, "GC %d SHRUNKEN %d CLAUSES\n", 
	 compsat->stats.gcs, shrunken_clauses);
  if (collected_literals)
    log (compsat, 2, "GC %d COLLECTED %d LITERALS\n", 
	 compsat->stats.gcs, collected_literals);
#endif
  compsat->stats.collected_literals += collected_literals;
  compsat->stats.collected_clauses += collected_clauses;

  /* 2. Flush references to garbage clauses, update new positions.
   */
  for (idx = 1; idx <= compsat->max; idx++)
    {
      for (sign = -1; sign <= 1; sign += 2)
	{
	  lit = sign * idx;
	  watched = get_watched (compsat, lit);
	  tmp = get_assignment (compsat, lit);
	  if (tmp)
	    {
	      count = count_stack (watched);
	      if (!count)
		continue;
#ifdef COMPSAT
	      log (compsat, 6, "GC %d DISCONNECTING ALL %d WATCHES FOR %d\n",
		   compsat->stats.gcs, count, lit);
#endif
	      release_stack (compsat, watched);
	    }
	  else
	    {
	      q = watched->start;
	      eow = watched->top;
	      for (p = q; p < eow; p++)
		{
		  tmp = *p;
		  assert (0 < tmp && tmp + literals < eol - 1);
		  pos = literals[tmp];
#ifdef COMPSAT_LOG
		  if (pos != tmp)
		    log_connect_with_lit (compsat, 6, "DISCONNECT", lit, tmp);
#endif
		  if (pos == GARBAGE)
		      continue;

		  literals[tmp] = lit;
		  *q++ = pos;		/* update new position */
		}
	      watched->top = q;
	    }
	}
    }

  /* 3. Compactify literals.
   */
  moved_literals = moved_clauses = 0;
  new_iteration = new_learned = 0;
  p = q = literals + 1;
  while (p < eol)
    {
      if (old_learned == p)
	new_learned = q;

      if (old_iteration == p)
	new_iteration = q;

      lit = *p++;
      if (lit == GARBAGE)
	{
	  assert (*p == GARBAGE);
	  p++;
	  while (*p++)
	    ;
	}
      else
	{
	  q[0] = lit;
	  q[1] = *p++;
#ifdef COMPSAT_LOG
	  if (q + 2 != p)
	    {
	      log_connect (compsat, 6, "RECONNECT", q - literals);
	      log_connect (compsat, 6, "RECONNECT", q + 1 - literals);
	    }
#endif
	  moved_literals += 2;
	  moved_clauses++;
	  q += 2;

	  while ((lit = *p++))
	    {
	      tmp = get_assignment (compsat, lit);
	      assert (tmp != TRUE);
	      if (tmp == FALSE)
		continue;

	      moved_literals++;
	      *q++ = lit;
	    }

	  *q++ = 0;
	}

      assert (!p[-1]);
      assert (!q[-1]);
    }
  assert (q == new_eol);
  compsat->literals.top = new_eol;
  if (!new_learned)
    new_learned = new_eol;
  if (!new_iteration)
    new_iteration = new_eol;
  compsat->iteration = new_iteration - literals;
  compsat->learned = new_learned - literals;
  compsat->satisfied = new_eol - literals;
#ifdef COMPSAT_LOG
  if (moved_clauses)
    log (compsat, 2, "GC %d MOVED %d CLAUSES\n", 
	 compsat->stats.gcs, moved_clauses);
  if (moved_literals)
    log (compsat, 2, "GC %d MOVED %d LITERALS\n",
	 compsat->stats.gcs, moved_literals);
  if (new_learned != old_learned)
    log (compsat, 3, "GC %d MOVED FIRST LEARNED CLAUSE FROM %d TO %d\n",
	 compsat->stats.gcs,
	 old_learned - literals, new_learned - literals);
  if (new_learned != old_learned)
    log (compsat, 3, "GC %d MOVED PREVIOUS ITERATION INDEX FROM %d TO %d\n",
	 compsat->stats.gcs,
	 old_iteration - literals, new_iteration - literals);
#endif
#ifdef COMPSAT_CHECK
  check (compsat);
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
gc (CompSat * compsat)
{
#ifdef DECOMPOSE
  if (!compsat->stats.decompositions)
    decompose (compsat);
#endif
  compsat->stats.gcs++;
#ifdef COMPSAT_LOG
  log (compsat, 2, "GC %d START\n", compsat->stats.gcs);
#endif
  gc_implies (compsat);
  gc_watched (compsat);
#ifdef COMPSAT_LOG
  log (compsat, 2, "GC %d END\n", compsat->stats.gcs);
#endif
}

/*------------------------------------------------------------------------*/

inline_static int
preprocess (CompSat * compsat)
{
  int sum_pure_literals, num_pure_literals, num_units;

  assert (!compsat->contains_empty_clause);

  num_units = count_stack (&compsat->units);
  if (num_units)
    log (compsat, 1, "found %d units in original formula\n", num_units);

  if (!assign_units (compsat))
    return 0;

  if (more_unprocessed_assignments (compsat) && deduce (compsat))
    return 0;

  if (!is_empty_stack (&compsat->trail))
    gc (compsat);

  sum_pure_literals = 0;
  for (;;)
    {
      num_pure_literals = assign_pure_literals (compsat);
      if (!num_pure_literals)
	break;

      sum_pure_literals += num_pure_literals;
      gc (compsat);
    }

  log (compsat, 1, "found %d pure literals\n", sum_pure_literals);

  log (compsat, 1, "preprocessing produced %d assignments\n", 
       count_stack (&compsat->trail));

#ifdef DECOMPOSE
  decompose (compsat);
  log (compsat, 1, 
       "preprocessing results in %d non trivial components\n",
       compsat->stats.non_trivial_components);
#endif
  ref2score (compsat);
  order (compsat);

  return 1;
}

/*------------------------------------------------------------------------*/

static void
check_all_unassigned (CompSat * compsat)
{
#ifdef COMPSAT_CHECK
  int i, tmp;

  for (i = 1; i <= compsat->max; i++)
    {
      if (get_level (compsat, i) > 0)
	{
	  tmp = get_assignment (compsat, i);
	  assert (tmp == 0);
	}
    }
#endif
}

/*------------------------------------------------------------------------*/

inline_static void
iteration (CompSat * compsat)
{
  int num_pure_literals, sum_pure_literals;

  assert (!compsat->level);
  compsat->stats.iterations++;
  reset_counter (&compsat->counters.restart);
#ifdef COMPSAT_LOG
  log (compsat, 2, "ITERATION %d START\n", compsat->stats.iterations);
#endif

  gc (compsat);
  sum_pure_literals = 0;
  for (;;)
    {
      num_pure_literals = assign_pure_literals (compsat);
      if (!num_pure_literals)
	break;

      sum_pure_literals += num_pure_literals;
      gc (compsat);
    }

#ifdef COMPSAT_LOG
  log (compsat, 2, "ITERATION %d FOUND %d PURE LITERALS\n",
       compsat->stats.iterations, sum_pure_literals);
  log (compsat, 2, "ITERATION %d FINISHED WITH %d TOP LEVEL ASSIGNMENTS\n",
       compsat->stats.iterations, compsat->stats.toplevel);
#endif
  compsat->iteration = count_stack (&compsat->literals);
#ifdef DECOMPOSE
  decompose (compsat);
#endif
  order (compsat);
}

/*------------------------------------------------------------------------*/

inline_static int
backtrack (CompSat * compsat, int conflict)
{
  int clause, assignment, level, lit;

  if (!compsat->level)
    return 0;

  analyze (compsat, conflict);
  clause = add (compsat);

  assignment = drive (compsat, clause);

  level = jump (compsat, clause);
  untrail (compsat, level);

  lit = reason2lit (compsat, assignment);
  assign (compsat, lit, assignment);

  if (done_counter (&compsat->counters.rescore))
    rescore (compsat);
  else if (done_counter (&compsat->counters.sort))
    sort (compsat);

  if (!compsat->restart_at_next_decision &&
      done_counter (&compsat->counters.restart))
    compsat->restart_at_next_decision = 1;

  return 1;
}

/*------------------------------------------------------------------------*/

inline_static void
restart (CompSat * compsat)
{
  assert (compsat->restart_at_next_decision);
  compsat->restart_at_next_decision = 0;
#ifdef COMPSAT_LOG
  log (compsat, 2, "%d RESTART %.0f\n",
       compsat->level, compsat->counters.restart.finished);
#endif
  untrail (compsat, 0);
  check_all_unassigned (compsat);

  order (compsat);
}

/*------------------------------------------------------------------------*/
#ifdef DECOMPOSE
/*------------------------------------------------------------------------*/

inline_static int
all_assigned_in_component (CompSat * compsat, int component)
{
  int idx, tmp;
  
  for (idx = 1; idx <= compsat->max; idx++)
    {
      tmp = deref_component (compsat, idx);
      if (tmp != component)
	continue;

      if (!get_assignment(compsat, idx))
	return 0;
    }

  return 1;
}

/*------------------------------------------------------------------------*/

inline_static void
save_assignments_of_component (CompSat * compsat, int component)
{
  int satisfied, idx, tmp;

  if (!component)
    return;

  satisfied = all_assigned_in_component (compsat, component);

  /* I am not 100% sure that this really works.
   */
  assert (satisfied);
  if (!satisfied)
    return;

  for (idx = 1; idx <= compsat->max; idx++)
    {
      tmp = deref_component (compsat, idx);
      if (tmp != component)
	continue;

#ifdef COMPSAT_LOG
      tmp = get_assignment (compsat, idx);
      assert (tmp);
      log (compsat, 3, "%d SAVING ASSIGNMENT %d\n",
	   compsat->level, tmp = TRUE ? idx : -idx);
#endif
      compsat->stats.saved++;
      set_level (compsat, idx, 0);
    }
}

/*------------------------------------------------------------------------*/

inline_static void
switch_component (CompSat * compsat, int new_active)
{
  assert (new_active != compsat->active);
  if (compsat->active)
    {
      compsat->stats.switches++;
#ifdef COMPSAT_LOG
      log (compsat, 2, "%d SWITCHING FROM COMPONENT %d TO %d\n",
	   compsat->level, compsat->active, new_active);
#endif
    }
  compsat->active = new_active;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

inline_static int
decide (CompSat * compsat)
{
  int lit;
#ifdef DECOMPOSE
  int component;
#endif

  if (done_counter (&compsat->counters.report))
    report (compsat);

  if (compsat->restart_at_next_decision)
    restart (compsat);

  compsat->stats.decisions++;
  compsat->stats.sum_levels_at_decision += compsat->level;

  if ((lit = berkmin (compsat)) || (lit = vsids (compsat)))
    {
#ifdef COMPSAT_LOG
      log (compsat, 3, "%d DECIDE %d\n", compsat->level, lit);
#endif
#ifdef DECOMPOSE
      component = deref_component (compsat, ABS (lit));
      if (compsat->active != component)
	{
	  save_assignments_of_component (compsat, compsat->active);
	  switch_component (compsat, component);
	}
#endif
      new_level (compsat, lit);
    }
  else
    {
      compsat->unassigned = count_stack (&compsat->order);
#ifdef COMPSAT_LOG
      log (compsat, 3, "%d ALLASSIGNED\n", compsat->level);
#endif
#ifdef COMPSAT_CHECK
      if (compsat->checker.enabled)
	check_is_satisfied (compsat);
#endif
    }

  return lit;
}

/*------------------------------------------------------------------------*/

inline_static int
reached_decision_limit (CompSat * compsat)
{
  if (compsat->limits.decisions < 0)
    return 0;
  else
    return compsat->stats.decisions > compsat->limits.decisions;
}

/*------------------------------------------------------------------------*/

inline_static int
reached_time_limit (CompSat * compsat)
{
  double limit;
  int res;

  limit = compsat->limits.time;
  if (limit < 0)
    return 0;

  update_time (compsat);
  res = (compsat->stats.time >= limit);

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
sat (CompSat * compsat)
{
  int conflict;

  if (compsat->contains_empty_clause)
    return COMPSAT_UNSATISFIABLE;

  if (!preprocess (compsat))
    return COMPSAT_UNSATISFIABLE;

  for (;;)
    {
      if (!more_unprocessed_assignments (compsat))
	{
	  if (!decide (compsat))
	    return COMPSAT_SATISFIABLE;

	  if (reached_decision_limit (compsat))
	    return COMPSAT_DECISION_OUT;

	  if (reached_time_limit (compsat))
	    return COMPSAT_TIME_OUT;
	}

      conflict = deduce (compsat);

      if (conflict)
	{
	  if (!backtrack (compsat, conflict))
	    return COMPSAT_UNSATISFIABLE;
	}
      else if (!compsat->level)
	iteration (compsat);
    }
}

/*------------------------------------------------------------------------*/

inline_static void
enter (CompSat * compsat)
{
  if (!compsat->internal++)
    compsat->entered = compsat_get_time ();
}

/*------------------------------------------------------------------------*/

inline_static void
leave (CompSat * compsat)
{
  assert (compsat->internal > 0);

  if (compsat->internal == 1)
    {
      compsat->stats.time += delta_time (compsat);
      compsat->entered = 0;
    }

  compsat->internal--;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

inline_static int
err_aux (CompSat * compsat, const char *fmt, va_list ap)
{
  FILE *file;

  file = log_file (compsat);
  fprintf (file, "*** compsat: ");
  vfprintf (file, fmt, ap);
  fputc ('\n', file);

  return COMPSAT_ABORT;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

inline_static int
err (CompSat * compsat, const char *fmt, ...)
{
  va_list ap;
  int res;

  va_start (ap, fmt);
  res = err_aux (compsat, fmt, ap);
  va_end (ap);

  return res;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

CompSat *
compsat_new (void)
{
  CompSat *res;

  res = (CompSat *) calloc (1, sizeof (*res));
  push_stack (res, &res->literals, 0);
  res->learned = 1;
  res->iteration = 1;

  res->stats.time = 0;
  res->stats.decisions = 0;
  res->stats.conflicts = 0;
  res->stats.assignments = 0;

  res->limits.time = -1;
  res->limits.decisions = -1;
  res->limits.bytes = -1;

  init_counter (&res->counters.sort, 256, 1, -1);
  init_counter (&res->counters.rescore, 256, 1, -1);
  init_counter (&res->counters.restart, 1024, 2, 8192);
  init_counter (&res->counters.report, 1000, 1, 999);

#ifdef COMPSAT_CHECK
  init_checker (&res->checker, res);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

void
compsat_set_leak_hook (CompSat * compsat, size_t * leaked_ptr)
{
  *leaked_ptr = 0;;
#ifdef COMPSAT_CHECK
  compsat->leaked_ptr = leaked_ptr;
#endif
}

/*------------------------------------------------------------------------*/

void
compsat_delete (CompSat * compsat)
{
  if (!compsat)
    return;

#ifdef COMPSAT_CHECK
  release_checker (&compsat->checker);
#endif

  if (compsat->log && compsat->close_log)
    fclose (compsat->log);

  if (compsat->out && compsat->close_out)
    fclose (compsat->out);

  release_stack (compsat, &compsat->literals);
  release_stack (compsat, &compsat->units);
  release_stack (compsat, &compsat->trail);
  release_stack (compsat, &compsat->control);
  release_stack (compsat, &compsat->order);
  release_stack (compsat, &compsat->clause);
  release_stack (compsat, &compsat->visited);
  release_stack (compsat, &compsat->stack);
  release_stack (compsat, &compsat->added_literals);
  delete_string (compsat, compsat->error);
  delete_string (compsat, compsat->name);

  release_variables (compsat);

  release_stack (compsat, &compsat->mark[0]);
  release_stack (compsat, &compsat->mark[1]);
  release_stack (compsat, &compsat->used);

#ifdef COMPSAT_CHECK
  if (compsat->leaked_ptr)
    *compsat->leaked_ptr = compsat->stats.bytes;
#endif

  free (compsat);
}

/*------------------------------------------------------------------------*/

inline_static int
count_implications (CompSat * compsat)
{
  int res, idx, sign, lit;
  Stack * implies;

  res = 0;
  for (idx = 1; idx <= compsat->max; idx++)
    for (sign = 1; sign >= 0; sign--)
      {
	lit = (sign ? -1 : 1) * idx;
	implies = get_implies (compsat, lit);
	res += count_stack (implies);
      }

  assert (!(res  & 1));

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
count_clauses (CompSat * compsat)
{
  IteratorStack (&compsat->literals, it, 1);
  int lit, res;

  res = 0;
  lit = next_stack (it);
  assert (!lit);

  while (!done_stack (it))
    {
      lit = next_stack (it);
      if (!lit)
	res++;
    }

  return res;
}

/*------------------------------------------------------------------------*/

inline_static void
print_clauses (CompSat * compsat, FILE * file)
{
  int lit;

  IteratorStack (&compsat->literals, litit, 1);
  lit = next_stack (litit);
  assert (!lit);
  while (!done_stack (litit))
    {
      lit = next_stack (litit);
      fprintf (file, "%d%c", lit, lit ? ' ' : '\n');
    }
}

/*------------------------------------------------------------------------*/

inline_static void
print_implications (CompSat * compsat, FILE * file)
{
  int idx, sign, lit, other, res;
  Stack * implies;

  res = 0;
  for (idx = 1; idx <= compsat->max; idx++)
    for (sign = 1; sign >= 0; sign--)
      {
	lit = (sign ? -1 : 1) * idx;
	implies = get_implies (compsat, lit);
	IteratorStack (implies, it, 1);
	while (!done_stack (it))
	  {
	    other = next_stack (it);
	    if (ABS(other) < ABS(lit))
	      continue;

	    fprintf (file, "%d %d 0\n", -lit, other);
	  }
      }
}

/*------------------------------------------------------------------------*/

inline_static void
print_units (CompSat * compsat, FILE * file)
{
  IteratorStack (&compsat->units, unitit, 1);
  while (!done_stack (unitit))
    fprintf (file, "%d 0\n", next_stack (unitit));
}

/*------------------------------------------------------------------------*/

void
compsat_print (CompSat * compsat, FILE * file)
{
  int count;

  enter (compsat);

  if (compsat->contains_empty_clause)
    fprintf (file, "p cnf 0 1\n0\n");
  else
    {
      count = count_clauses (compsat);
      count += count_implications (compsat) / 2;
      count += count_stack (&compsat->units);

      fprintf (file, "p cnf %d %d\n", compsat->max, count);

      print_clauses (compsat, file);
      print_implications (compsat, file);
      print_units (compsat, file);
    }

  leave (compsat);
}

/*------------------------------------------------------------------------*/

inline_static void
compsat_add_clause (CompSat * compsat, const int *clause)
{
  int clause_idx;

  enter (compsat);

  insert_variables (compsat, clause);
  clause_idx = simplify_clause (compsat, clause);

  if (clause_idx)
    {
      order_clause (compsat, clause_idx);
      register_literals (compsat, clause_idx, 0);
      connect_clause (compsat, clause_idx, 0);
    }

  leave (compsat);
}

/*------------------------------------------------------------------------*/

void
compsat_add_literal (CompSat * compsat, int literal)
{
  enter (compsat);

  push_stack (compsat, &compsat->added_literals, literal);
  if (!literal)
    {
      compsat_add_clause (compsat, compsat->added_literals.start);
      reset_stack (compsat, &compsat->added_literals, 0);
    }

  leave (compsat);
}

/*------------------------------------------------------------------------*/

const char *
compsat_read (CompSat * compsat, const char *name)
{
  int added_literals, added_clauses, added_variables;
  int old_literals, old_variables;
  double time_before_read, delta;
  const char *error;
  const int *clause;
  Dimacs *parser;
  DimacsMem mem;

  enter (compsat);

  memset (&mem, 0, sizeof (mem));
  mem.mgr = compsat;
  mem.new = (DimacsNew) new;
  mem.delete = (DimacsDelete) delete;

  log (compsat, 1, "reading %s\n", name);
  time_before_read = compsat_get_time ();
  old_literals = compsat->stats.literals;
  old_variables = compsat->stats.variables;

  /* Since trivial clauses are not really added we have to count the number
   * of clauses found in the file.
   */
  added_clauses = 0;
  parser = dimacs_new (name, &mem);
  while ((clause = dimacs_read (parser)))
    {
      compsat_add_clause (compsat, clause);
      added_clauses++;
    }

  /* Variables of trivial clauses are still added.    For literals this is
   * not the case but we just don't care about literals in trivial clauses.
   */
  added_literals = compsat->stats.literals - old_literals;
  added_variables = compsat->stats.variables - old_variables;

  error = dimacs_error (parser);
  if (!error)
    {
      if (dimacs_specified_num_clauses (parser) < added_clauses)
	error = "more clauses found than specified";

      if (dimacs_specified_num_clauses (parser) > added_clauses)
	error = "less clauses found than specified";

      /* TODO: if we want to allow multiple read we have to fix this test.
       */
      if (dimacs_specified_max_variable (parser) < compsat->max)
	error = "specified maximal variable index exceeded";
    }

  if (error)
    {
      delete_string (compsat, compsat->error);
      compsat->error = copy_string (compsat, error);
    }
  else
    {
      delta = compsat_get_time () - time_before_read;
      delta = (delta < 0) ? 0 : delta;

      log (compsat, 1, "read %d literals in %0.2f seconds\n",
	   added_literals, delta);

      log (compsat, 1, "found %d new variables in %d clauses\n",
	   added_variables, added_clauses);
    }

#ifdef COMPSAT_CHECK
  if (!error)
    check (compsat);
#endif

  dimacs_delete (parser);
  leave (compsat);

  return compsat->error;
}

/*------------------------------------------------------------------------*/

size_t
compsat_bytes (CompSat * compsat)
{
  return compsat->stats.max_bytes + sizeof (*compsat);
}

/*------------------------------------------------------------------------*/

double
compsat_seconds (CompSat * compsat)
{
  if (compsat->internal)
    update_time (compsat);

  return compsat->stats.time;
}

/*------------------------------------------------------------------------*/

void
compsat_adjust_timer (CompSat * compsat)
{
  update_time (compsat);
}

/*------------------------------------------------------------------------*/

void
compsat_print_seconds (FILE * file, double seconds)
{
  fprintf (file, "%.2f sec", seconds);
}

/*------------------------------------------------------------------------*/

void
compsat_print_bytes (FILE * file, size_t bytes)
{
  if (bytes < (1 << 10))
    fprintf (file, "%.0f bytes", (double) bytes);
  else if (bytes < (1 << 20))
    fprintf (file, "%.1f KB", ((double) bytes) / (double) (1 << 10));
  else if (bytes < (1 << 30))
    fprintf (file, "%.1f MB", ((double) bytes) / (double) (1 << 20));
  else
    fprintf (file, "%.1f GB", ((double) bytes) / (double) (1 << 30));
}

/*------------------------------------------------------------------------*/

FILE *
compsat_out_file (CompSat * compsat)
{
  return out_file (compsat);
}

/*------------------------------------------------------------------------*/

FILE *
compsat_log_file (CompSat * compsat)
{
  return log_file (compsat);
}

/*------------------------------------------------------------------------*/

FILE *
compsat_err_file (CompSat * compsat)
{
  return log_file (compsat);	/* no explicit error file */
}

/*------------------------------------------------------------------------*/

void
compsat_stats (CompSat * compsat, FILE * file)
{
  enter (compsat);

  fprintf (file, COMPSAT_PREFIX
	   "%d units, %d pure literals, %d zombies\n",
	   compsat->stats.units,
	   compsat->stats.pure_literals,
	   compsat->stats.zombies);

#ifdef DECOMPOSE
  fprintf (file, COMPSAT_PREFIX
	   "%d non trivial components, %d maximum\n",
	   compsat->stats.non_trivial_components,
	   compsat->stats.max_non_trivial_components);
#endif

  fprintf (file, COMPSAT_PREFIX
	   "%d iterations, %d gcs"
#ifdef DECOMPOSE
	   ", %d decompositions"
	   ", %d switches"
#endif
	   "\n",
	   compsat->stats.iterations,
	   compsat->stats.gcs
#ifdef DECOMPOSE
	   , compsat->stats.decompositions,
	   compsat->stats.switches
#endif
	   );

#ifdef DECOMPOSE
  fprintf (file, COMPSAT_PREFIX
           "%.2f sec decomposition time, "
	   "%d cross component collected clauses\n",
	   compsat->stats.decomposition_time,
	   compsat->stats.cross_component_collected_learned_clauses);
#endif

  fprintf (file, COMPSAT_PREFIX
	   "%.0f assignments, %d at top level"
#ifdef DECOMPOSE
	   ", %d saved"
#endif
	   "\n",
	   compsat->stats.assignments,
	   compsat->stats.toplevel
#ifdef DECOMPOSE
	   , compsat->stats.saved
#endif
	   );

  fprintf (file, COMPSAT_PREFIX
	   "%.0f conflicts, %.0f decisions, %.1f average level\n",
	   compsat->stats.conflicts,
	   compsat->stats.decisions,
	   avglevel (compsat));

  fprintf (file, COMPSAT_PREFIX
           "%.0f osearched, %.0f csearched, %.0f lsearched\n",
	   compsat->stats.osearched,
	   compsat->stats.csearched,
	   compsat->stats.lsearched);

  fprintf (file, COMPSAT_PREFIX
	   "%.0f rescores, %.0f sortings, %.0f restarts\n",
	   compsat->counters.rescore.finished,
	   compsat->counters.sort.finished,
	   compsat->counters.restart.finished);

  fprintf (file, COMPSAT_PREFIX
	   "collected %.0f implications, %.0f clauses, %.0f literals\n",
	   compsat->stats.collected_implications,
	   compsat->stats.collected_clauses,
	   compsat->stats.collected_literals);

  fprintf (file, COMPSAT_PREFIX
	   "%d variables, %d implications, %d clauses, %d literals\n",
	   compsat->stats.variables,
	   compsat->stats.implications,
	   compsat->stats.clauses,
	   compsat->stats.literals);

  fputs (COMPSAT_PREFIX, file);
  compsat_print_seconds (file, compsat_seconds (compsat));
  fputs (", ", file);
  compsat_print_bytes (file, compsat_bytes (compsat));
  fputc ('\n', file);

  leave (compsat);
}

/*------------------------------------------------------------------------*/

const char *
compsat_res2str (int res)
{
  switch (res)
    {
    case COMPSAT_SATISFIABLE:
      return ("SATISFIABLE");
      break;
    case COMPSAT_UNSATISFIABLE:
      return ("UNSATISFIABLE");
      break;
    case COMPSAT_TIME_OUT:
      return ("TIMEOUT");
      break;
    case COMPSAT_SPACE_OUT:
      return ("SPACEOUT");
      break;
    case COMPSAT_DECISION_OUT:
      return ("DECISIONOUT");
      break;
    case COMPSAT_EXIT:
      return ("EXIT");
      break;
    default:
      assert (res == COMPSAT_ABORT);
      return ("ABORT");
      break;
    }
}

/*------------------------------------------------------------------------*/

int
compsat_sat (CompSat * compsat)
{
  int res;

  enter (compsat);
  res = sat (compsat);
  if (compsat->stats.reported > 0)
    {
      report (compsat);
      log (compsat, 0, "\n");
    }
  leave (compsat);

  return res;
}

/*------------------------------------------------------------------------*/

int
compsat_deref (CompSat * compsat, int literal)
{
  int res;

  if (literal < 0)
    return -compsat_deref (compsat, -literal);

  if (literal > compsat->max)
    return 0;

  res = get_assignment (compsat, literal);
  res = (res == TRUE);

  return res;
}

/*------------------------------------------------------------------------*/

#define LINE_WIDTH 80
#define MAX_POS 77

void
compsat_print_assignment (CompSat * compsat, FILE * file)
{
  char buffer[LINE_WIDTH], str[20];
  int i, tmp, pos, len;

  assert (LINE_WIDTH > MAX_POS);

  pos = 0;

  for (i = 1; i <= compsat->max + 1; i++)
    {
      if (i <= compsat->max)
	{
	  tmp = compsat_deref (compsat, i); 
	  tmp = tmp ? i : -i;
	}
      else
	tmp = 0;		/* SAT 2003 competition format */

      /* TODO: this could be slow.
       */
      sprintf (str, "%d", tmp);
      len = strlen (str);

      if (pos + len >= MAX_POS)
	{
	  buffer[pos] = 0;
	  fputs ("v ", file);	/* SAT 2003 competition format */
	  fputs (buffer, file);
	  fputc ('\n', file);
	  pos = 0;
	}

      if (pos)
	buffer[pos++] = ' ';

      strcpy (buffer + pos, str);
      pos += len;
    }

  if (pos)
    {
      fputs ("v ", file);	/* SAT 2003 competition format */
      fputs (buffer, file);
      fputc ('\n', file);
    }
}

/*------------------------------------------------------------------------*/


inline_static int
is_number (const char *str)
{
  const char *p;
  int res;

  p = str;

  if (*p == '-')
    p++;

  if (isdigit ((int) *p++))
    {
      while (isdigit ((int) *p))
	p++;

      res = !*p;
    }
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static char
character_class (char ch)
{
  if (ch >= '0' && ch <= '9')
    return '0';

  return ch;
}

/*------------------------------------------------------------------------*/

inline_static int
is_float (const char *str)
{
  const char *p;
  int digits;

  digits = 0;
  p = str;

  goto L1;

L1:				/* -?(0*.0+ | 0+) */

  switch (character_class (*p++))
    {
    case '-':
      goto L2;

    case '.':
      goto L3;

    case '0':
      goto L4;
    }

  return 0;

L2:				/* 0*.0+ | 0+ */

  switch (character_class (*p++))
    {
    case '0':
      goto L4;

    case '.':
      goto L3;
    }

  return 0;

L3:				/* 0+ */

  switch (character_class (*p++))
    {
    case '0':
      goto L5;
    }

  return 0;

L4:				/* 0*.0+ | 0* */

  switch (character_class (*p++))
    {
    case '0':
      goto L4;

    case '.':
      goto L3;

    case 0:
      return 1;
    }

  return 0;

L5:				/* 0* */

  switch (character_class (*p++))
    {
    case '0':
      goto L5;

    case 0:
      return 1;
    }

  return 0;
}

/*------------------------------------------------------------------------*/

CompSatArgvIterator *
compsat_new_ArgvIterator (CompSat * compsat, int argc, char **argv)
{
  CompSatArgvIteratorInternal *internal;
  CompSatArgvIterator *res;

  internal = new (compsat, sizeof (*internal));
  internal->compsat = compsat;
  internal->argc = argc;
  internal->argv = argv;
  internal->pos = 1;
  internal->opt_buffer[0] = '-';

  res = new (compsat, sizeof (*res));
  res->internal = internal;

  return res;
}

/*------------------------------------------------------------------------*/

inline_static int
compsat_is_option_with_argument (const char *opt)
{
  assert (opt[0] == '-');
  assert (opt[2] == 0);

  switch (opt[1])
    {
    case 'l':
    case 'o':
    case 't':
    case 'd':
    case 'm':
#ifdef COMPSAT_WHITE
    case 'W':
#endif
      return 1;
    default:
      return 0;
    }
}

/*------------------------------------------------------------------------*/

int
compsat_is_file_option (const char *opt)
{
  if (opt[0] == '-' && !opt[2])
    return (opt[1] == 'l' || opt[1] == 'o');

  return !strcmp (opt, "--divert");
}

/*------------------------------------------------------------------------*/

void
compsat_report_invalid_option (CompSat * compsat,
			       CompSatArgvIterator * it, const char *err)
{
  FILE *file;

  file = compsat_err_file (compsat);

  fprintf (file, "*** '%s", it->opt);
  if (it->arg)
    fprintf (file, "[ ]%s", it->arg);
  fprintf (file, "': %s\n", err);
  fflush (file);
}

/*------------------------------------------------------------------------*/

int
compsat_has_more_elements_ArgvIterator (CompSatArgvIterator * it,
					const char **err_msg)
{
  char *str;

  *err_msg = 0;

  if (it->internal->pos >= it->internal->argc)
    return 0;

  str = it->internal->argv[it->internal->pos++];
  it->opt = str;

  if (str[0] == '-' && str[1] && str[1] != '-')
    {
      if (str[2])
	{
	  it->arg = str + 2;
	  it->internal->opt_buffer[1] = str[1];
	  it->opt = it->internal->opt_buffer;
	}
      else if (!compsat_is_option_with_argument (str))
	it->arg = 0;
      else if (it->internal->pos < it->internal->argc)
	it->arg = it->internal->argv[it->internal->pos++];
      else
	{
	  *err_msg = "argument is missing";
	  return 0;
	}
    }
  else
    it->arg = 0;

  return 1;
}

/*------------------------------------------------------------------------*/

void
compsat_delete_ArgvIterator (CompSatArgvIterator * it)
{
  CompSat *compsat;

  compsat = it->internal->compsat;
  delete (compsat, it->internal, sizeof (*it->internal));
  delete (compsat, it, sizeof (*it));
}

/*------------------------------------------------------------------------*/

#ifdef COMPSAT_CHECK
#define CHECK_OPTION \
"  --check            enable expensive checks\n"
#else
#define CHECK_OPTION ""
#endif

#define USAGE \
"  -v                 increase verbosity level\n" \
"  -V                 set maximal verbosity\n" \
"  -o[ ]<out-file>    use <out-file> for printing (default stdout)\n" \
"  -l[ ]<log-file>    use <log-file> for logging (default stderr)\n" \
"  --divert           divert logging to <out-file>\n" \
CHECK_OPTION \
"\n" \
"  -d[ ]<num>         max decisions (default -1 = unrestricted)\n" \
"  -m[ ]<num>         max mega bytes (default -1 = unrestricted)\n" \
"  -t[ ]<num>         max seconds (default -1 = unrestricted)\n"

/*------------------------------------------------------------------------*/

void
compsat_set_verbose_level (CompSat * compsat, int verbosity)
{
  if (compsat->verbosity == verbosity)
    return;

  compsat->verbosity = verbosity;
  compsat->logged_verbosity = 0;
}

/*------------------------------------------------------------------------*/

void
compsat_set_out_file (CompSat * compsat, FILE * file)
{
  if (compsat->out && compsat->close_out)
    fclose (compsat->out);

  compsat->out = file;
  compsat->close_out = 0;
}

/*------------------------------------------------------------------------*/

void
compsat_set_log_file (CompSat * compsat, FILE * file)
{
  if (compsat->log && compsat->close_log)
    fclose (compsat->log);

  compsat->log = file;
  compsat->close_log = 0;
}

/*------------------------------------------------------------------------*/

int
compsat_get_verbose_level (CompSat * compsat)
{
  return compsat->verbosity;
}

/*------------------------------------------------------------------------*/

const char *
compsat_prefix (CompSat * compsat)
{
  (void) compsat;
  return COMPSAT_PREFIX;
}

/*------------------------------------------------------------------------*/

const char *
compsat_usage (void)
{
  return USAGE;
}

/*------------------------------------------------------------------------*/

static const char *could_not_open_file = "could not open file";
static const char *argument_is_missing = "argument is missing";
static const char *argument_is_no_integer = "argument is no integer";
static const char *argument_is_no_float = "argument is no float";
static const char *invalid_option = "invalid option";

/*------------------------------------------------------------------------*/

const char *
compsat_option (CompSat * compsat, const char *opt, const char *arg)
{
  double number;

  if (!strcmp (opt, "-o"))
    {
      if (!arg)
	return argument_is_missing;

      if (compsat->out && compsat->close_out)
	fclose (compsat->out);

      compsat->out = fopen (arg, "w");
      if (!compsat->out)
	return could_not_open_file;

      compsat->close_out = 1;
    }
  else if (!strcmp (opt, "-l"))
    {
      if (!arg)
	return argument_is_missing;

      if (compsat->log && compsat->close_log)
	fclose (compsat->log);

      compsat->log = fopen (arg, "w");
      if (!compsat->log)
	return could_not_open_file;
      compsat->close_log = 1;
    }
#ifdef COMPSAT_CHECK
  else if (!strcmp (opt, "--check"))
    {
      compsat->checker.enabled = 1;
    }
#endif
  else if (!strcmp (opt, "--divert"))
    {
      compsat->divert = 1;
    }
  else if (!strcmp (opt, "-v"))
    {
      if (compsat->verbosity < INT_MAX)
	compsat->verbosity++;
    }
  else if (!strcmp (opt, "-V"))
    {
      compsat->verbosity = INT_MAX;
    }
  else if (!strcmp (opt, "-m"))
    {
      if (!arg)
	return argument_is_missing;

      if (!is_float (arg))
	return argument_is_no_float;

      number = strtod (arg, NULL);
      if (number >= (LONG_MAX >> 20))
	return 0;

      if (number < 0)
	compsat->limits.bytes = -1;
      else
	compsat->limits.bytes = (number * (1 << 20));
    }
  else if (!strcmp (opt, "-d"))
    {
      if (!arg)
	return argument_is_missing;

      if (!is_number (arg))
	return argument_is_no_integer;

      number = strtod (arg, NULL);
      compsat->limits.decisions = (number < 0) ? -1 : number;
    }
  else if (!strcmp (opt, "-t"))
    {
      if (!arg)
	return argument_is_missing;

      if (!is_float (arg))
	return argument_is_no_float;

      number = strtod (arg, NULL);
      compsat->limits.time = (number < 0) ? -1 : number;

      /* TODO: do we want to add the already used up time here ? */
    }
  else if (opt[0] == '-')
    return invalid_option;
  else
    compsat->name = copy_string (compsat, opt);

  return 0;
}

/*------------------------------------------------------------------------*/

const char *
compsat_id (void)
{
  return COMPSAT_ID;
}

/*------------------------------------------------------------------------*/

const char *
compsat_copyright (void)
{
  return COMPSAT_COPYRIGHT;
}

/*------------------------------------------------------------------------*/

#define CONFIGURATION \
COMPSAT_CC " " COMPSAT_CFLAGS "\n" \
COMPSAT_OS "\n" 

/*------------------------------------------------------------------------*/

const char *
compsat_configuration (void)
{
  return CONFIGURATION;
}

/*------------------------------------------------------------------------*/

const char *
compsat_version (void)
{
  return COMPSAT_VERSION;
}

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

static void
white_white (CompSat * compsat)
{
  out (compsat, "this is the output of the white-box test case\n");
}

/*------------------------------------------------------------------------*/

static void
white_stack (CompSat * compsat)
{
  Stack stack;
  int i;

  memset (&stack, 0, sizeof (stack));

  reset_stack (compsat, &stack, 0);

  for (i = 0; i < 2000; i++)
    push_stack (compsat, &stack, i);

  for (i = 0; i < 1000; i++)
    (void) pop_stack (compsat, &stack);

  reset_stack (compsat, &stack, 500);

  set_stack (compsat, &stack, 499, -499);
  set_stack (compsat, &stack, 500, -500);
  set_stack (compsat, &stack, 501, -501);
  set_stack (compsat, &stack, 511, -511);
  set_stack (compsat, &stack, 513, -513);

  while (!is_empty_stack (&stack))
    out (compsat, "%d\n", pop_stack (compsat, &stack));

  release_stack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_pointer0 (CompSat * compsat)
{
  Stack stack;
  int tmp;

  memset (&stack, 0, sizeof (stack));
  push_stack (compsat, &stack, 0);
  push_stack (compsat, &stack, 1);
  push_stack (compsat, &stack, 2);
  push_stack (compsat, &stack, 3);
  push_stack (compsat, &stack, 4);
  {
    CursorStack (&stack, it, 2, -1);
    while ((tmp = next_stack (it)))
      out (compsat, "%d\n", tmp);
  }
  out (compsat, "---------------------------------\n");
  {
    CursorStack (&stack, it, 1, 1);
    while ((tmp = next_stack (it)) != 4)
      out (compsat, "%d\n", tmp);
  }
  out (compsat, "---------------------------------\n");
  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }
  out (compsat, "---------------------------------\n");
  reset_stack (compsat, &stack, 0);
  push_stack (compsat, &stack, 5);
  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }
  out (compsat, "---------------------------------\n");
  release_stack (compsat, &stack);
  push_stack (compsat, &stack, 6);
  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }
  release_stack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_pointer1 (CompSat * compsat)
{
  Stack stack;
  void *s[2];
  int tmp;

  memset (&stack, 0, sizeof (stack));
  push_stack (compsat, &stack, 0);
  push_stack (compsat, &stack, 1);
  push_stack (compsat, &stack, 2);

  {
    CursorStack (&stack, it, 0, 1);
    save_stack_cursor (it, s[0]);
    (void) next_stack (it);
    save_stack_cursor (it, s[1]);
    (void) next_stack (it);
    swap_stack (&stack, it, s[0], s[1]);
  }

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }

  out (compsat, "---------------------------------\n");

  {
    CursorStack (&stack, it, 2, -1);
    save_stack_cursor (it, s[0]);
    (void) next_stack (it);
    save_stack_cursor (it, s[1]);
    tmp = next_stack (it);
    assert (tmp == 0);
    swap_stack (&stack, it, s[0], s[1]);
  }

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }

  release_stack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_compact (CompSat * compsat)
{
  Stack stack;
  memset (&stack, 0, sizeof (stack));
  enlarge_stack (compsat, &stack);
  log (compsat, 0, "%d\n", size_stack (&stack));
  release_stack (compsat, &stack);
  log (compsat, 0, "%d\n", size_stack (&stack));
  enlarge_stack (compsat, &stack);
  log (compsat, 0, "%d\n", size_stack (&stack));
  enlarge_stack (compsat, &stack);
  log (compsat, 0, "%d\n", size_stack (&stack));
  reset_stack (compsat, &stack, 0);
  log (compsat, 0, "%d\n", size_stack (&stack));
  enlarge_stack (compsat, &stack);
  log (compsat, 0, "%d\n", size_stack (&stack));
  enlarge_stack (compsat, &stack);
  log (compsat, 0, "%d\n", size_stack (&stack));
  release_stack (compsat, &stack);
  log (compsat, 0, "%d\n", size_stack (&stack));
}

/*------------------------------------------------------------------------*/

static void
white_compact0 (CompSat * compsat)
{
  Stack stack;
  int elem;

  memset (&stack, 0, sizeof (stack));
  push_stack (compsat, &stack, 0);
  push_stack (compsat, &stack, 1);
  push_stack (compsat, &stack, 2);
  push_stack (compsat, &stack, 3);
  push_stack (compsat, &stack, 4);

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      {
	elem = next_stack (it);
	log (compsat, 0, "%d %d\n", elem, size_stack (&stack));
      }
  }

  log (compsat, 0, "------------------------------------\n");

  push_stack (compsat, &stack, 5);

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      {
	elem = next_stack (it);
	log (compsat, 0, "%d %d\n", elem, size_stack (&stack));
      }
  }

  log (compsat, 0, "------------------------------------\n");

  reset_stack (compsat, &stack, 0);

  push_stack (compsat, &stack, 1000);
  push_stack (compsat, &stack, 1001);
  push_stack (compsat, &stack, 1002);
  push_stack (compsat, &stack, 1003);
  push_stack (compsat, &stack, 1004);
  push_stack (compsat, &stack, 1005);
  push_stack (compsat, &stack, 65536);
  push_stack (compsat, &stack, 1007);
  push_stack (compsat, &stack, 1008);
  push_stack (compsat, &stack, 1009);

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      {
	elem = next_stack (it);
	log (compsat, 0, "%d %d\n", elem, size_stack (&stack));
      }
  }

  log (compsat, 0, "------------------------------------\n");

  push_stack (compsat, &stack, 2000);
  push_stack (compsat, &stack, 2001);
  push_stack (compsat, &stack, 2002);
  push_stack (compsat, &stack, 2003);

  {
    IteratorStack (&stack, it, -1);
    while (!done_stack (it))
      {
	elem = next_stack (it);
	log (compsat, 0, "%d %d\n", elem, size_stack (&stack));
      }
  }

  reset_stack (compsat, &stack, 0);

  {
    IteratorStack (&stack, it, -1);
    assert (done_stack (it) || next_stack (it));
  }
  {
    IteratorStack (&stack, it, 1);
    assert (done_stack (it) || next_stack (it));
  }

  release_stack (compsat, &stack);
  {
    IteratorStack (&stack, it, 1);
    assert (done_stack (it) || next_stack (it));
  }
  {
    IteratorStack (&stack, it, -1);
    assert (done_stack (it) || next_stack (it));
  }
}

/*------------------------------------------------------------------------*/

static void
white_compact1_line (CompSat * compsat, Stack * stack)
{
  out (compsat, "-- count=%d -----------\n", count_stack (stack));
}

/*------------------------------------------------------------------------*/

static void
white_compact1 (CompSat * compsat)
{
  Stack stack;
  int tmp, i;

  memset (&stack, 0, sizeof (stack));
  for (i = 0; i < 2; i++)
    {
      push_stack (compsat, &stack, 0);
      push_stack (compsat, &stack, 1);
      push_stack (compsat, &stack, 2);
      push_stack (compsat, &stack, 3);
      push_stack (compsat, &stack, 4);
      if (i)
	push_stack (compsat, &stack, 1 << 16);

      {
	CursorStack (&stack, it, 2, -1);
	while ((tmp = next_stack (it)))
	  out (compsat, "%d\n", tmp);
      }
      white_compact1_line (compsat, &stack);
      {
	CursorStack (&stack, it, 1, 1);
	while ((tmp = next_stack (it)) != 4)
	  out (compsat, "%d\n", tmp);
      }
      white_compact1_line (compsat, &stack);
      {
	IteratorStack (&stack, it, 1);
	while (!done_stack (it))
	  out (compsat, "%d\n", next_stack (it));
      }
      white_compact1_line (compsat, &stack);
      {
	IteratorStack (&stack, it, -1);
	while (!done_stack (it))
	  out (compsat, "%d\n", next_stack (it));
      }
      white_compact1_line (compsat, &stack);
      reset_stack (compsat, &stack, 0);
      push_stack (compsat, &stack, 5);
      {
	IteratorStack (&stack, it, 1);
	while (!done_stack (it))
	  out (compsat, "%d\n", next_stack (it));
      }
      white_compact1_line (compsat, &stack);
      release_stack (compsat, &stack);
      push_stack (compsat, &stack, 6);
      {
	IteratorStack (&stack, it, 1);
	while (!done_stack (it))
	  out (compsat, "%d\n", next_stack (it));
      }
      release_stack (compsat, &stack);
      if (!i)
	white_compact1_line (compsat, &stack);
    }
}

/*------------------------------------------------------------------------*/

static void
white_compact2 (CompSat * compsat)
{
  Stack stack;

  memset (&stack, 0, sizeof (stack));
  push_stack (compsat, &stack, 0);
  push_stack (compsat, &stack, 1);
  push_stack (compsat, &stack, 2);
  push_stack (compsat, &stack, 3);
  push_stack (compsat, &stack, 4);

  out (compsat, "%d\n", get_stack (&stack, 1));
  out (compsat, "%d\n", get_stack (&stack, 4));
  push_stack (compsat, &stack, 1 << 16);
  out (compsat, "%d\n", get_stack (&stack, 1));
  out (compsat, "%d\n", get_stack (&stack, 4));

  release_stack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_compact3 (CompSat * compsat)
{
  Stack stack;
  void *s[2];
  int tmp;

  memset (&stack, 0, sizeof (stack));
  push_stack (compsat, &stack, 0);
  push_stack (compsat, &stack, 1);
  push_stack (compsat, &stack, 2);

  {
    CursorStack (&stack, it, 0, 1);
    save_stack_cursor (it, s[0]);
    (void) next_stack (it);
    save_stack_cursor (it, s[1]);
    (void) next_stack (it);
    swap_stack (&stack, it, s[0], s[1]);
  }

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }

  out (compsat, "---------------------------------\n");

  {
    CursorStack (&stack, it, 2, -1);
    save_stack_cursor (it, s[0]);
    (void) next_stack (it);
    save_stack_cursor (it, s[1]);
    tmp = next_stack (it);
    assert (tmp == 0);
    swap_stack (&stack, it, s[0], s[1]);
  }

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }

  push_stack (compsat, &stack, 1 << 16);

  {
    CursorStack (&stack, it, 3, -1);
    save_stack_cursor (it, s[0]);
    (void) next_stack (it);
    save_stack_cursor (it, s[1]);
    tmp = next_stack (it);
    assert (tmp == 0);
    swap_stack (&stack, it, s[0], s[1]);
  }

  out (compsat, "---------------------------------\n");

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }

  release_stack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

static void
white_compact4 (CompSat * compsat)
{
  Stack stack;

  memset (&stack, 0, sizeof (stack));

  push_stack (compsat, &stack, 32767);
  push_stack (compsat, &stack, -32768);
  push_stack (compsat, &stack, 32768);

  {
    IteratorStack (&stack, it, 1);
    while (!done_stack (it))
      out (compsat, "%d\n", next_stack (it));
  }

  release_stack (compsat, &stack);
}

/*------------------------------------------------------------------------*/

#define N_WHITE_BITMAP 1000

static void
white_bitmap (CompSat * compsat)
{
  int i;

#ifdef USE_BITMAP
  Stack bitmap;

  memset (&bitmap, 0, sizeof (bitmap));

  for (i = 0; i < N_WHITE_BITMAP; i += 3)
    set_bit (compsat, &bitmap, i);

  for (i = 0; i < N_WHITE_BITMAP; i += 2)
    clr_bit (compsat, &bitmap, i);

  for (i = 0; i < N_WHITE_BITMAP; i++)
    if (get_bit (compsat, &bitmap, i))
      out (compsat, "%d\n", i);

  release_stack (compsat, &bitmap);
#else
  for (i = 0; i < N_WHITE_BITMAP; i++)
    if (!(i % 3) && (i % 2))
      out (compsat, "%d\n", i);
#endif
}

/*------------------------------------------------------------------------*/

#define N_WHITE_PPLONG 1000
#define M_WHITE_PPLONG 10

static void
white_pplong (CompSat * compsat)
{
  char name[200];
  int i, j, r, k;
  FILE *file;
  char ch;

  sprintf (name, "/tmp/white-pplong-%u", (unsigned) getpid ());
  file = fopen (name, "w");
  if (!file)
    return;

  fprintf (file, "p cnf %d %d\n", N_WHITE_PPLONG - 1, M_WHITE_PPLONG);

  r = 13;

  for (k = 0; k < M_WHITE_PPLONG; k++)
    {
      for (i = 1; i <= N_WHITE_PPLONG; i++)
	{
	  if (i < N_WHITE_PPLONG)
	    {
	      fprintf (file, "%d", (!(r % 2) ? -1 : 1) * i);
	      r = (101 * r + 17) % 113;
	    }
	  else
	    fputc ('0', file);

	  assert (r >= 0);

	  for (j = 0; j <= (i % 5); j++)
	    {
	      if (!((i + j) % 17))
		ch = '\t';
	      else if (!((i + j) % 31))
		ch = '\n';
	      else
		ch = ' ';

	      fputc (ch, file);
	    }
	}
    }

  fclose (file);

  compsat_read (compsat, name);
  if (compsat->error)
    err (compsat, "%s", compsat->error);
  else
    out (compsat, "%d %d\n", compsat->max, count_clauses (compsat));

  unlink (name);
}

/*------------------------------------------------------------------------*/

#define N_WHITE_LARGE ((1 << 20) - 1)

static void
white_large (CompSat * compsat)
{
  char name[200];
  FILE *file;

  sprintf (name, "/tmp/white-large-%u", (unsigned) getpid ());
  file = fopen (name, "w");
  if (!file)
    return;

  fprintf (file, "p cnf %d 1\n%d 0\n", N_WHITE_LARGE, N_WHITE_LARGE);
  fclose (file);

  compsat_read (compsat, name);
  if (compsat->error)
    err (compsat, "%s", compsat->error);
  else
    out (compsat, "%d %d\n", compsat->max, count_clauses (compsat));

  unlink (name);
}

/*------------------------------------------------------------------------*/

static void
white_memout (CompSat * compsat)
{
  int res;

#ifdef COMPSAT_WHITE
  void *ptr;

  ptr = new (compsat, 13);
  compsat->force_out_of_memory = 1;
  res = !new (compsat, 17);
  if (res)
    {
      res = !resize (compsat, ptr, 13, 200);
      if (res)
	delete (compsat, ptr, 13);
    }
#else
  res = 0;
#endif

  out (compsat, "%d\n", res);
}

/*------------------------------------------------------------------------*/

static void
white_memzero (CompSat * compsat)
{
  out (compsat, "%.0f\n", (double) (unsigned long) new (compsat, 0));
  out (compsat, "%.0f\n",
       (double) (unsigned long) resize (compsat, new (compsat, 17), 17, 0));
}

/*------------------------------------------------------------------------*/

static void
white_isfloat_aux (CompSat * compsat, const char *str)
{
  out (compsat, "\"%s\" %d\n", str, is_float (str));
}

/*------------------------------------------------------------------------*/

static void
white_isfloat (CompSat * compsat)
{
  white_isfloat_aux (compsat, "0");
  white_isfloat_aux (compsat, ".0");
  white_isfloat_aux (compsat, "3.1415");
  white_isfloat_aux (compsat, "-1");
  white_isfloat_aux (compsat, "-1.23");
  white_isfloat_aux (compsat, "-.2");
  white_isfloat_aux (compsat, "47.11");
  white_isfloat_aux (compsat, "-47.11");

  white_isfloat_aux (compsat, "");
  white_isfloat_aux (compsat, "a");
  white_isfloat_aux (compsat, ".");
  white_isfloat_aux (compsat, "0.");
  white_isfloat_aux (compsat, "-a");
  white_isfloat_aux (compsat, "0a");
  white_isfloat_aux (compsat, "0.0a");
}

/*------------------------------------------------------------------------*/

static void
white_dimacs0 (CompSat * compsat)
{
  FILE *file = log_file (compsat);
  const char *p;

  for (p = dimacs_id (); *p; p++)
    {
      switch (*p)
	{
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case '.':
	case '$':
	  break;

	default:
	  fputc (*p, file);
	  break;
	}
    }

  fputc ('\n', file);
}

/*------------------------------------------------------------------------*/

static void
white_log_from_file (CompSat * compsat, const char *name)
{
  FILE *file;
  int ch;

  file = fopen (name, "r");
  if (file)
    {
      while ((ch = fgetc (file)) != EOF)
	log (compsat, 0, "%c\n", ch);

      fclose (file);
    }
}

/*------------------------------------------------------------------------*/

#define LOG_STR "/tmp/compsat-white-log"

static void
white_log (CompSat * compsat)
{
  CompSat *other;
  other = compsat_new ();
  compsat_option (other, "-l", LOG_STR);
  log (other, 0, "logged\n");
  compsat_delete (other);
  white_log_from_file (compsat, LOG_STR);
  unlink (LOG_STR);
}

/*------------------------------------------------------------------------*/

#define OUT_STR "/tmp/compsat-white-out"

static void
white_out (CompSat * compsat)
{
  CompSat *other;
  other = compsat_new ();
  compsat_option (other, "-o", OUT_STR);
  out (other, "printed\n");
  compsat_option (other, "--divert", 0);
  log (other, 0, "logged again\n");
  compsat_delete (other);
  white_log_from_file (compsat, OUT_STR);
  unlink (OUT_STR);
}

/*------------------------------------------------------------------------*/

static void
white_cmp (CompSat * compsat)
{
  int res, clause[3];

  clause[0] = 1;
  clause[1] = 2;
  clause[2] = 0;
  compsat_add_clause (compsat, clause);
  clause[0] = -1;
  clause[1] = 0;
  compsat_add_clause (compsat, clause);

  res = 1;

  if (res)
    res = (cmp (compsat, 1, 1) == 0);
  if (res)
    res = (cmp (compsat, -1, -1) == 0);

  if (res)
    res = (cmp (compsat, 1, -1) < 0);
  if (res)
    res = (cmp (compsat, -1, 1) > 0);

  if (res)
    res = (cmp (compsat, 1, 2) > 0);
  if (res)
    res = (cmp (compsat, 2, 1) < 0);

  set_score (compsat, 2, 1);

  if (res)
    res = (cmp (compsat, 1, 2) < 0);
  if (res)
    res = (cmp (compsat, 2, 1) > 0);

  set_score (compsat, 1, 1);

  if (res)
    res = (cmp (compsat, 1, 2) > 0);
  if (res)
    res = (cmp (compsat, 2, 1) < 0);

  log (compsat, 0, "%d\n", res);
}

/*------------------------------------------------------------------------*/

#define N_SORT 111 * 111	/* has to be square number */

static void
white_sort (CompSat * compsat)
{
  int *clause, i, j, delta, count, sorted;

  clause = new (compsat, sizeof (clause[0]) * (N_SORT + 1));
  for (delta = 0; delta * delta < N_SORT; delta++)
    ;
  assert (delta * delta == N_SORT);
  count = 1;
  for (i = 0; i < delta; i++)
    for (j = 0; j < delta; j++)
      clause[j * delta + i] = count++;
  clause[N_SORT] = 0;
  compsat_add_clause (compsat, clause);

  hypersort (compsat, clause, N_SORT);

  sorted = 1;
  for (i = 0; sorted && i < N_SORT - 1; i++)
    sorted = (cmp (compsat, clause[i], clause[i + 1]) > 0);

  delete (compsat, clause, sizeof (clause[0]) * (N_SORT + 1));

  log (compsat, 0, "%s\n", sorted ? "sorted" : "unsorted");
}

/*------------------------------------------------------------------------*/

typedef void (*TestCase) (CompSat *);

#define TC(pattern) \
do { if(!strcmp(# pattern, name)) return white_ ## pattern; } while(0)

static TestCase
get_white_box_test_case (const char *name)
{
  TC (white);
  TC (stack);
  TC (pointer0);
  TC (pointer1);
  TC (compact);
  TC (compact0);
  TC (compact1);
  TC (compact2);
  TC (compact3);
  TC (compact4);
  TC (bitmap);
  TC (pplong);
  TC (large);
  TC (memout);
  TC (memzero);
  TC (isfloat);
  TC (dimacs0);
  TC (log);
  TC (out);
  TC (cmp);
  TC (sort);

  return 0;
}

/*------------------------------------------------------------------------*/

int
compsat_is_white_box_test_case (const char *name)
{
  return get_white_box_test_case (name) != 0;
}

/*------------------------------------------------------------------------*/

void
compsat_run_white_box_test_case (CompSat * compsat, const char *name)
{
  TestCase tc;

  tc = get_white_box_test_case (name);
  assert (tc);
  tc (compsat);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
